@@ -27,7 +27,7 @@ extern int sh_read(), sh_echo(), sh_read(), sh_test(), sh_include();
 extern int sh_cd(), sh_colon(), sh_hash(), sh_bce(), sh_eval(), sh_export();
 extern int sh_getopts(), sh_shift(), sh_umask(), sh_set(), sh_unset();
 extern int sh_wait(), sh_times(), sh_trap(), sh_type(), sh_builtin();
-extern int sh_sleep();
+extern int sh_sleep(), sh_expr();
 
 
 extern char *ifs;
@@ -65,33 +65,34 @@ struct shCmd builtins[] = {
 {	"setf",		0,		sh_setf,	SH_ARGV		},
 {	"get",		0,		sh_get,		SH_ARGV		},
 {	"length",	0,		sh_length,	SH_ARGV		},
-{	"[",		sh_test,	0,		0,		},
-{	"test",		sh_test,	0,		0,		},
-{	"echo",		sh_echo,	0,		0,		},
-{	"cd",		sh_cd,		0,		0,		},
-{	"hash",		sh_hash,	0,		0,		},
-{	"read",		sh_read,	0,		0,		},
-{	":",		sh_colon,	0,		0,		},
-{	".",		sh_include,	0,		0,		},
-{	"break",	sh_bce,		0,		SH_INTERNAL,	},
-{	"continue",	sh_bce,		0,		SH_INTERNAL,	},
-{	"return",	0,		sh_return,	SH_ARGV,	},
-{	"returns",	0,		sh_returns,	SH_ARGV,	},
-{	"exit",		sh_bce,		0,		SH_INTERNAL,	},
-{	"eval",		sh_eval,	0,		0,		},
-{	"export",	sh_export,	0,		0,		},
-{	"getopts",	sh_getopts,	0,		0,		},
-{	"shift",	sh_shift,	0,		0,		},
-{	"umask",	sh_umask,	0,		0,		},
-{	"set",		sh_set,		0,		0,		},
-{	"unset",	sh_unset,	0,		0,		},
-{	"wait",		sh_wait,	0,		0,		},
-{	"times",	sh_times,	0,		0,		},
-{	"trap",		sh_trap,	0,		0,		},
-{	"type",		sh_type,	0,		0,		},
-{	"builtin",	sh_builtin,	0,		0,		},
-{	"sleep",	sh_sleep,	0,		0,		},
-{	0,		0,		0,		0,		},
+{	"[",		sh_test,	0,		0		},
+{	"test",		sh_test,	0,		0		},
+{	"echo",		sh_echo,	0,		0		},
+{	"cd",		sh_cd,		0,		0		},
+{	"hash",		sh_hash,	0,		0		},
+{	"read",		sh_read,	0,		0		},
+{	":",		sh_colon,	0,		0		},
+{	".",		sh_include,	0,		0		},
+{	"break",	sh_bce,		0,		SH_INTERNAL	},
+{	"continue",	sh_bce,		0,		SH_INTERNAL	},
+{	"return",	0,		sh_return,	SH_ARGV		},
+{	"returns",	0,		sh_returns,	SH_ARGV		},
+{	"exit",		sh_bce,		0,		SH_INTERNAL	},
+{	"expr",		sh_expr,	0,		SH_ARGV		},
+{	"eval",		sh_eval,	0,		0		},
+{	"export",	sh_export,	0,		0		},
+{	"getopts",	sh_getopts,	0,		0		},
+{	"shift",	sh_shift,	0,		0		},
+{	"umask",	sh_umask,	0,		0		},
+{	"set",		sh_set,		0,		0		},
+{	"unset",	sh_unset,	0,		0		},
+{	"wait",		sh_wait,	0,		0		},
+{	"times",	sh_times,	0,		0		},
+{	"trap",		sh_trap,	0,		0		},
+{	"type",		sh_type,	0,		0		},
+{	"builtin",	sh_builtin,	0,		0		},
+{	"sleep",	sh_sleep,	0,		0		},
+{	0,		0,		0,		0		},
 };
 
 struct conscell *
@@ -1125,3 +1126,24 @@ sh_sleep(argc, argv)
 	return 0;
 }
 
+int
+sh_expr(argc,argv)
+	int argc;
+	char *argv[];
+{
+#ifndef	MAILER
+	fprintf(stderr, "%s: Not implemented w/o Zmailer\n",argv[0]);
+	return 1;
+#else
+	regexp *re = NULL;
+
+	if (argc != 4 || argv[2][0] != ':') {
+		fprintf(stderr, USAGE_EXPR, argv[0]);
+		return 1;
+	}
+/* XX: expr  code! */
+	re = NULL;
+	fprintf(stderr,"%s: Not yet implemented!\n",argv[0]);
+	return 1;
+#endif
+}
