@@ -261,7 +261,7 @@ siomore(siop)
 	  } else {
 	    /* copy existing buffer into malloc'ed ditto */
 	    siop->sb_base = (u_char *)emalloc(siop->sb_bufsiz);
-	    bcopy(prevbase, siop->sb_base, siop->sb_ptr - prevbase);
+	    memcpy(siop->sb_base, prevbase, siop->sb_ptr - prevbase);
 	    siop->sb_flag |= O_CREAT;
 	  }
 	  siop->sb_ptr += siop->sb_base - prevbase;
@@ -644,7 +644,7 @@ pforw:
 			/* the string or number proper */
 			if (siop->sb_cnt - (n = size) >= 0) {
 				siop->sb_cnt -= n;
-				bcopy(t, (char *)siop->sb_ptr, n);
+				memcpy((char *)siop->sb_ptr, t, n);
 				siop->sb_ptr += n;
 			} else
 				while (--n >= 0)
