@@ -1,6 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Some functions Copyright 1991-1994 Matti Aarnio.
  */
 
 /*
@@ -34,18 +35,19 @@
 static int run_hostname ARGCV;
 static int run_whataddress ARGCV;
 static int run_erraddrlog ARGCV;
+
 static struct conscell *run_dblookup __((struct conscell *avl, struct conscell *il));
 extern struct conscell *sh_car();
 static struct conscell *run_cadr __((struct conscell *avl, struct conscell *il));
 static struct conscell *run_caddr __((struct conscell *avl, struct conscell *il));
 static struct conscell *run_cadddr __((struct conscell *avl, struct conscell *il));
+static struct conscell *run_listexpand __((struct conscell *avl, struct conscell *il));
 
 static int run_stability ARGCV;
 static int run_process ARGCV;
 static int run_grpmems ARGCV;
 static int run_praliases ARGCV;
 static int run_listaddrs ARGCV;
-static int run_listexpand ARGCV;
 static int run_homedir ARGCV;
 static int run_822date ARGCV;
 static int run_filepriv ARGCV;
@@ -100,7 +102,7 @@ struct shCmd fnctns[] = {
 {	"groupmembers",	run_grpmems,		0,		0	},
 {	"printaliases",	run_praliases,		0,		0	},
 {	"listaddresses",run_listaddrs,		0,		0	},
-{	"listexpand",	run_listexpand,		0,		SH_ARGV	},
+{	"listexpand",	0,			run_listexpand,	SH_ARGV	},
 {	"homedirectory",run_homedir,		0,		0	},
 {	"rfc822date",	run_822date,		0,		0	},
 {	"filepriv",	run_filepriv,		0,		0	},
@@ -692,7 +694,7 @@ rd_stability(dirp,dirs)
 			continue;
 		}
 		dearray[deindex].f_name = nbindex;
-		bcopy(dp->d_name, nbarray + nbindex, namelen+1);
+		memcpy(nbarray + nbindex, dp->d_name, namelen+1);
 		nbindex += namelen + 1;
 
 		++deindex;
@@ -830,14 +832,28 @@ run_daemon(argc, argv)
 	i = 0;
 	for (;;) {
 		/*
-		 * We do a seekdir() instead of opendir()/closedir()
+		 * We would like to do a rewinddir() instead of
+		 * opendir()/closedir()
 		 * inside the loop, to avoid allocating and freeing
 		 * an 8k chunk of memory all the time.  This can lead
 		 * to memory fragmentation and thus growing VM.
+		 *
+		 * However several systems do NOT guarantee that
+		 * rewinddir() will find any new data from the
+		 * system...
 		 */
-		/* We ASSUME that  seekdir(fd,0L)  is the proper thing
-		   to do. */
-		seekdir(dirp[i], 0L); /* essentially a REWIND().. */
+		/* rewinddir(dirp[i]); /* some system w/o this ? */
+
+#ifdef	BUGGY_CLOSEDIR
+		/*
+		 * Major serious bug time here;  some closedir()'s
+		 * free dirp before referring to dirp->dd_fd. GRRR.
+		 * XX: remove this when bug is eradicated from System V's.
+		 */
+		close(dirp[i]->dd_fd);
+#endif
+		closedir(dirp[i]);
+		dirp[i] = opendir(dirs[i]);
 
 		if (stability)
 			did_cnt = rd_stability(dirp[i],dirs[i]);
@@ -1171,63 +1187,84 @@ FILE *fp;
 }
 #endif
 
-static int
-run_listexpand(argc,argv)
-	int argc;
-	char *argv[];
+/* listexpand()  -- do a bunch of tasks usually done with a group of
+   functions, starting with  "listaddresses" ... */
+
+static struct conscell *
+run_listexpand(avl, il)
+	struct conscell *avl, *il;
 {
 	struct header hs;
 	struct envelope *e;
 	struct address *ap;
 	struct token **pt, *t;
 	struct addr *pp;
-	int c, n, errflag, stuff;
+	struct conscell *attributes, *tmp;
+	struct conscell *al, *alp;
+	char *localpart, *origaddr;
+	int c, n, len, errflag, stuff, cnt;
 	char *comment, *erroraddress;
 	u_char *s, *s_end;
 	struct siobuf *osiop;
 	FILE *mfp, *fp, f;
 	char buf[4096];
 	int fd2;
+	char *olderrors = errors_to;
+
+	extern struct conscell *s_value;
+
+	il = cdar(avl); /* The CDR (next) of the arg list.. */
 
 	errflag = 0;
 	erroraddress = NULL;
 	comment = "list";
 	optind = 1;
 
-	while ((c = getopt(argc, argv, "c:e:E:")) != EOF) {
-		switch (c) {
-		  case 'c':
-		    comment = optarg;
-		    break;
-		  case 'e':
-		    erroraddress = optarg;
-		    break;
-		  case 'E':
-		    if (errors_to) free(errors_to);
-		    errors_to = (void*)malloc(strlen(optarg)+1);
-		    if (errors_to)
-		      strcpy(errors_to, optarg);
-		    break;
-		  default:
-		    ++errflag;
-		    break;
-		}
+	while (il != NULL && STRING(il) && il->string[0] == '-' &&
+	       cdr(il) != NULL && STRING(cdr(il))) {
+	  switch( il->string[1] ) {
+	    case 'c':
+	      comment = cdr(il)->string;
+	      break;
+	    case 'e':
+	      erroraddress = cdr(il)->string;
+	      break;
+	    case 'E':
+	      errors_to = cdr(il)->string;
+	      break;
+	    default:
+	      errflag = 1;
+	      break;
+	  }
+	  il = cddr(il); /* Skip TWO entries at the time! */
 	}
-	if (errflag) {
+	cnt = 0;
+	tmp = il;
+	for (; tmp != NULL; tmp = cdr(tmp)) ++cnt; /* Count arguments.. */
+
+	if (errflag || cnt != 3) {
 		fprintf(stderr,
 			"Usage: %s [ -e error-address ] [ -E errors-to-address ] [ -c comment ] $attribute $localpart $origaddr < /file/path \n",
-			argv[0]);
-		return 1;
+			car(avl)->string);
+		errors_to = olderrors;
+		return NULL;
 	}
-	/* We (memopry-)leak this stuff for a moment.. */
+
+	attributes = copycell(il);
+	car(attributes) = s_copy_tree(car(attributes));
+	cdr(attributes) = NULL;
+	localpart  = cdr(il)->string;
+	origaddr   = cddr(il)->string;
+
+	/* We (memory-)leak this stuff for a moment.. */
 	e = (struct envelope *)tmalloc(sizeof (struct envelope));
 	e->e_nowtime = now;
-	e->e_file = argv[0];
+	e->e_file = car(avl)->string;
 	/* we only need sensible `e' if its a time stamp header,
 	   which it ain't, so "listaddress" is just fine for e_file .. */
 
 	hs.h_descriptor = &aliashdr;
-	hs.h_pname = argv[0];
+	hs.h_pname = car(avl)->string;
 
 	/* create h->h_lines */
 	/*
@@ -1242,11 +1279,12 @@ run_listexpand(argc,argv)
 	 * the fake I/O stuff in ../libsh.
 	 */
 	if ((fp = fdopen(0, "r")) == NULL) {
-		fprintf(stderr, "%s: fdopen failed\n", argv[0]);
+		fprintf(stderr, "%s: fdopen failed\n", car(avl)->string);
 #ifdef DEBUG_FOPEN
 		report_fds(stderr);
 #endif
-		return 1;
+		errors_to = olderrors;
+		return NULL;
 	}
 	fd2 = dup(0);	/* Copy the file handle for latter returning
 			   AFTER an fclose() has closed the primary
@@ -1317,6 +1355,7 @@ run_listexpand(argc,argv)
 	if (c && (*s == ',' || *s == '\n'))
 		*s = '\0';
 	*pt = NULL;
+	s = buf; /* For debugger uses far below... */
 
 	hs.h_contents = hdr_scanparse(e, &hs, 1);
 	hs.h_stamp = hdr_type(&hs);
@@ -1328,6 +1367,8 @@ run_listexpand(argc,argv)
 		osiop = siofds[fileno(mfp)];
 		siofds[fileno(mfp)] = NULL;
 		fprintf(mfp, "channel error\n");
+		fprintf(mfp, "to %s\n", erroraddress);
+		fprintf(mfp, "From: Error Channel <>\n");
 		fprintf(mfp, "To: %s\n", erroraddress);
 		fprintf(mfp, "Subject: Error in %s\n\n", comment);
 		hdr_errprint(e, &hs, mfp, comment);
@@ -1340,28 +1381,132 @@ run_listexpand(argc,argv)
 	  } else {		/* if (hs.h_contents.a == NULL) */
 	    if (erroraddress != NULL)
 	      printf("%s\n", erroraddress);
-	    fprintf(stderr, "%s: null input\n", argv[0]);
+	    fprintf(stderr, "%s: null input\n", car(avl)->string);
 	  }
-	  return 1;
+	  errors_to = olderrors;
+	  return NULL;
 	}
+	cnt = 0;
 	for (ap = hs.h_contents.a; ap != NULL; ap = ap->a_next) {
+		char *av[5];
+		int rc;
+		struct conscell *l, *d;
+
 		for (pp = ap->a_tokens; pp != NULL; pp = pp->p_next)
 			if (pp->p_type == anAddress)
 				break;
 		if (pp == NULL)
 			continue;
-		pureAddress(stdout, ap->a_tokens);
-		/* printAddress(stdout, ap->a_tokens, 0); */
-		putc('\n', stdout);
+		/* pureAddress(stdout, ap->a_tokens); */
+		/** printAddress(stdout, ap->a_tokens, 0); **/
+		/* putc('\n', stdout); */
+		buf[0] = 0;
+		len = printToken(buf,buf+sizeof(buf),ap->a_tokens->p_tokens, NULL, 0);
+		if (buf[0] == 0) continue; /* Burp ??? */
+
+#define cdddr(x) cdr(cddr(x))
+
+		/* XX: We leak these at the moment! */
+		l       = newstring(strsave(buf));
+		cdr(l)  = conststring(origaddr);
+		cddr(l) = car(attributes);
+		l = ncons(l);
+
+		deferit = 0;
+		v_set(DEFER, "");
+
+		rc = l_apply("rrouter", l);
+
+		if (deferit && (d = v_find(DEFER))) {
+		  s_free_tree(s_value);
+		  s_value = NULL;
+		  l = conststring((u_char *)"hold");
+		  cdr(l)   = copycell(cdr(d));
+		  cddr(l)  = newstring(strsave(buf));
+		  cdddr(l) = car(attributes);
+		  l = ncons(l);
+		  l = ncons(l);
+		} else 	if (rc != 0 || s_value == NULL || !LIST(s_value)) {
+		  /* $(rrouter xx xx xx)  returned something invalid.. */
+		  s_free_tree(s_value);
+		  s_value == NULL;
+		  continue;
+		} else {
+		  /*
+		   * We expect router to either return
+		   * (local - user attributes) or (((local - user attributes)))
+		   * or ( ((local - user attr)) ((local - another attr)) )
+		   */
+		  if (LIST(car(s_value))) {
+		    if (!LIST(caar(s_value)) || !STRING(caaar(s_value))) {
+		      fprintf(stderr,
+			      "%s: '%s' returned invalid 2-level list: ",
+			      progname, ROUTER);
+		      s_grind(s_value, stderr);
+		      s_free_tree(s_value);
+		      s_value = NULL;
+		      errors_to = olderrors;
+		      return NULL;
+		    }
+		    l = s_copy_tree(s_value);
+		  } else {
+		    l = s_copy_tree(s_value);
+		    l = ncons(l);
+		    l = ncons(l);
+		  }
+
+		  s_free_tree(s_value);
+		  s_value = NULL;
+		}
+
+		/* Now the  "(struct conscell *) l" contains a route,
+		   it is a time to put them together.. */
+		/* l ->  ( ((chan param param2 attrs)) )  */
+		if (al == NULL) {
+		  al  = l;		/* The head anchor */
+		  alp = car(al);	/* address list END node ptr */
+		  while (cdr(alp) != NULL) {
+		    alp = cdr(alp);
+		    ++cnt;
+		  }
+		} else {
+		  /*
+		     (
+		      ((chan param param2 attrs))
+		      ((chan param param2 attrs))
+		     )
+		   */
+		  cdr(alp) = car(l);	/* Attach the new list to the old */
+		  while (cdr(alp) != NULL) {
+		    alp = cdr(alp);	/* .. and traverse to its end ... */
+		    ++cnt;
+		  }
+		}
+
+		/* Debugging.. */
+		/* buf[len] = '\n';
+		   buf[len+1] = 0;
+		   fwrite(buf,1,len+1,stdout); */
 	}
-	/*
-	 * XX: If the alias expansion was a mail group,
-	 * we didn't print anything.
-	 */
-	return 0;
+
+	&cnt; /* For debugging.. */
+
+	if (al == NULL) { /* ERROR! NO ADDRESSES! */
+	  al = conststring((u_char *)"error");
+	  cdr(al)  = conststring((u_char *)"expansion");
+	  cddr(al) = newstring(strsave(localpart));
+	  al = ncons(al);
+	  al = ncons(al);
+	  errors_to = olderrors;
+	  return al;
+	}
+
+	errors_to = olderrors;
+	return al;
 
 } /* end-of: run_listexpand() */
 
+
 static int
 run_listaddrs(argc, argv)
 	int argc;
@@ -1379,6 +1524,7 @@ run_listaddrs(argc, argv)
 	FILE *mfp, *fp, f;
 	char buf[4096];
 	int fd2;
+	char *old_errorsto = errors_to;
 
 	errflag = 0;
 	erroraddress = NULL;
@@ -1394,10 +1540,7 @@ run_listaddrs(argc, argv)
 			erroraddress = optarg;
 			break;
 		case 'E':
-			if (errors_to) free(errors_to);
-			errors_to = (void*)malloc(strlen(optarg)+1);
-			if (errors_to)
-			  strcpy(errors_to, optarg);
+			errors_to = (void*)strsave(optarg);
 			break;
 		default:
 			++errflag;
@@ -1518,6 +1661,8 @@ run_listaddrs(argc, argv)
 		osiop = siofds[fileno(mfp)];
 		siofds[fileno(mfp)] = NULL;
 		fprintf(mfp, "channel error\n");
+		fprintf(mfp, "to %s\n", erroraddress);
+		fprintf(mfp, "From: Error Channel <>\n");
 		fprintf(mfp, "To: %s\n", erroraddress);
 		fprintf(mfp, "Subject: Error in %s\n\n", comment);
 		hdr_errprint(e, &hs, mfp, comment);
@@ -1599,6 +1744,7 @@ run_filepriv(argc, argv)
 	int argc;
 	char *argv[];
 {
+	char *argv0 = argv[0];
 	int id;
 	struct stat stbuf;
 	char *file, *cp, *dir;
@@ -1613,7 +1759,9 @@ run_filepriv(argc, argv)
 	  --argc;
 	}
 	if (argc == 1 || argc > 3 || (maxperm & (~0664)) != 0) {
-		fprintf(stderr, "Usage: %s [-M maxperm] pathname [ uid ]\n", argv[0]);
+		fprintf(stderr, "Usage: %s [-M maxperm] pathname [ uid ]\n", argv0);
+		if (maxperm & (~0664))
+		  fprintf(stderr, "       maxperm must be 664 or stricter!\n");
 		return 1;
 	}
 	file = argv[1];
@@ -1624,12 +1772,12 @@ run_filepriv(argc, argv)
 			/* if we can't open it, don't trust it */
 			perror(file);
 			fprintf(stderr,
-				"%s: cannot fopen(\"%s\")!\n", argv[0], file);
+				"%s: cannot fopen(\"%s\")!\n", argv0, file);
 			return 2;
 		}
 		if (fstat(fileno(fp), &stbuf) < 0) {
 			fprintf(stderr, "%s: cannot fstat(\"%s\")!\n",
-				argv[0], file);
+				argv0, file);
 			fclose(fp);
 			return 3;
 		}
@@ -1655,12 +1803,12 @@ run_filepriv(argc, argv)
 #else
 	dir = emalloc(cp - file + 1);
 #endif
-	bcopy(file, dir, cp - file);
+	memcpy(dir, file, cp - file);
 	dir[cp - file] = '\0';
 
 	if (stat(dir, &stbuf) < 0 || !S_ISDIR(stbuf.st_mode)) {
 		fprintf(stderr, "%s: not a directory: \"%s\"!\n",
-			argv[0], dir);
+			argv0, dir);
 #ifndef	USE_ALLOCA
 		free(dir);
 #endif
