@@ -24,6 +24,11 @@
 #include <resolv.h>
 #include "search.h"
 
+#ifdef	__alpha		/* The IP-address processing must be changed to
+			   something truly portable, this is awful.. */
+# define u_long u_int
+#endif
+
 /*
  * Query a BIND standard (version 4.7 or later) nameserver.
  */
@@ -62,19 +67,28 @@ typedef char msgdata;
    get returned to us, when we make a query with  T_ANY ... */
 
 #ifndef	T_TXT
-# define T_TXT 16
+# define T_TXT 16	/* Text strings */
 #endif
 #ifndef T_RP
-# define T_RP 17
+# define T_RP 17	/* Responsible person */
 #endif
 #ifndef T_AFSDB
-# define T_AFSDB 18
+# define T_AFSDB 18	/* AFS cell database */
+#endif
+#ifndef T_X25
+# define T_X25 19	/* X.25 calling address */
+#endif
+#ifndef T_ISDN
+# define T_ISDN 20	/* ISDN calling address */
+#endif
+#ifndef T_RT
+# define T_RT 21	/* router */
 #endif
 #ifndef T_NSAP
-# define T_NSAP 22
+# define T_NSAP 22	/* NSAP address */
 #endif
 #ifndef T_NSAP_PTR
-# define T_NSAP_PTR 23
+# define T_NSAP_PTR 23	/* reverse NSAP lookup (depreciated) */
 #endif
 #ifndef	T_UINFO
 # define T_UINFO 100
@@ -89,7 +103,7 @@ typedef char msgdata;
 # define T_UNSPEC 103
 #endif
 #ifndef T_SA
-# define T_SA 200
+# define T_SA 200		/* Shuffle addresses */
 #endif
 
 
@@ -353,7 +367,7 @@ getmxrr(host, localhost, ttlp, depth)
 		type = _getshort(cp);
  		cp += sizeof(u_short);				/* type */
  		cp += sizeof(u_short);				/* class */
-		ttl = _getlong(cp);
+		ttl = (time_t) _getlong(cp);
 		cp += sizeof(u_long);				/* ttl */
 		n = _getshort(cp);
 		cp += sizeof(u_short);				/* dlen */
@@ -581,7 +595,7 @@ getrrtype(host, rrtype, ttlp, depth)
 		type = _getshort(cp);
  		cp += sizeof(u_short);				/* type */
  		cp += sizeof(u_short);				/* class */
-		ttl = _getlong(cp);
+		ttl = (time_t) _getlong(cp);
 		cp += sizeof(u_long);				/* ttl */
 		n = _getshort(cp);
 		cp += sizeof(u_short);				/* dlen */
@@ -640,7 +654,7 @@ getrrtype(host, rrtype, ttlp, depth)
 			if (rrtype == T_ANY) {
 			    return newstring(strsave(host));
 			}
-			bcopy(cp, (char *)&inaddr, sizeof inaddr);
+			memcpy((char *)&inaddr, cp, sizeof inaddr);
 			s = dottedquad(&inaddr);
 			s = strsave(s);
 			if (lhead == NULL)
@@ -683,10 +697,14 @@ getrrtype(host, rrtype, ttlp, depth)
 			*ttlp = maxttl;
 			return newstring(strsave(nbuf));
 
+	/*	case T_AAAA:	IPv6 'A'-record */
 		case T_SOA:
 		case T_NS:
 		case T_RP:
 		case T_AFSDB:
+		case T_X25:
+		case T_ISDN:
+		case T_RT:
 		case T_NSAP:
 		case T_NSAP_PTR:
 		case T_HINFO:
