@@ -57,7 +57,7 @@ struct headerinfo optional_hdrs[] = {
 
 struct headerinfo envelope_hdrs[] = {
 #ifdef	USE_AUTH
-{ "authinfo",		AMailboxList,	nilUserType,	eAuthinfo	},
+{ "authinfo",	nilHeaderSemantics,	nilUserType,	eAuthinfo	},
 #endif
 { "channel",		AnyWord,	nilUserType,	eChannel	},
 { "envid",	nilHeaderSemantics,	nilUserType,	eEnvid		},
