@@ -57,6 +57,7 @@ extern void initline();
 
 static void	prctladdr __((struct conscell *info, FILE *fp, int cfflag, char *comment));
 static void	prdsndata __((struct conscell *info, FILE *fp, int cfflag, char *comment));
+static struct conscell	*find_errto __((struct conscell *list));
 
 /*
  * Apply RFC822 parsing and processing to the message in the file in argv[1].
@@ -867,12 +868,14 @@ struct rwmatrix {
 	struct rwmatrix	*down;
 	struct conscell *info;	/* rewrite, sender, or recipient list info */
 	struct notary   *dsn;
+	struct conscell *errto;
 	union {
 		int	number;	/* XOR address id number for recipients */
 		struct header *h;	/* for rewritings */
 	} urw;
 };
 
+
 static struct rwmatrix *
 rwalloc(rwpp)
 	struct rwmatrix **rwpp;
@@ -882,6 +885,7 @@ rwalloc(rwpp)
 	p = (struct rwmatrix *)tmalloc(sizeof (struct rwmatrix));
 	p->down = NULL;
 	p->dsn  = NULL;
+	p->errto = NULL;
 	p->urw.number = 0;			/* unused but why not */
 	if (*rwpp != NULL)
 		p->next = *rwpp;
@@ -1001,6 +1005,8 @@ sequencer(e, file)
 	char *envid = NULL;
 
 
+	errors_to = NULL; /* to be gotten rid off.. */
+
 	if (e == NULL)
 		return PERR_OK;
 	dprintf("Parse envelope and message header\n");
@@ -1384,13 +1390,15 @@ sequencer(e, file)
 		}
 	}
 #ifdef notdef
-	rewrite all addresses in the message according to the incoming-rewriting
-		rules for the originating channel.
+	rewrite all addresses in the message according to the
+	incoming-rewriting rules for the originating channel.
 #endif
 	dprintf("Route recipient addresses\n");
 	routed_addresses = NULL;
-	if (errors_to) free(errors_to);
+
+	/* if (errors_to) free(errors_to); */
 	errors_to = NULL;
+
 	for (h = e->e_eHeaders; h != NULL; h = h->h_next) {
 		static struct notary *DSN;
 		if (h->h_descriptor->class != eTo &&
@@ -1451,7 +1459,7 @@ sequencer(e, file)
 		/* first level is ANDs */
 		++idnumber, nxor = 0;
 		for (to = car(l); to != NULL; to = cdr(to)) {
-			struct conscell *x, *tmp;
+			struct conscell *x, *tmp, *errto;
 
 			/* secondlevel is XORs */
 #if 0
@@ -1474,11 +1482,14 @@ sequencer(e, file)
 			tmp = copycell(car(x));
 			cdr(tmp) = NULL;
 			for (rwp = rwhead; rwp != NULL; rwp = rwp->next)
-				if (s_equal(rwp->info, tmp))
+				if (s_equal(rwp->info, tmp) &&
+				    s_equal(rwp->errto,
+					    (errto = find_errto(cdar(x)))))
 					break;
 			if (rwp == NULL) {
 				rwp = rwalloc(&rwhead);
 				rwp->info = tmp;	/* rewritings */
+				rwp->errto = errto;
 			}
 			tmp = cddar(x);
 			cddar(x) = NULL;
@@ -1595,14 +1606,7 @@ sequencer(e, file)
 	 */
 
 	if (!iserrmessage()) {
-		if (errors_to != NULL) {	/* [mea@utu.fi] Stupid,
-						   but workable.. */
-			putc(_CF_ERRORADDR, ofp);
-			putc(_CFTAG_NORMAL, ofp);
-			fprintf(ofp,"%s\n",errors_to);
-			free(errors_to);
-			errors_to = NULL;
-		} else if ((h = erraddress(e)) != NULL) {
+		if ((h = erraddress(e)) != NULL) {
 			putc(_CF_ERRORADDR, ofp);
 			putc(_CFTAG_NORMAL, ofp);
 			for (ap = h->h_contents.a;
@@ -1634,6 +1638,23 @@ sequencer(e, file)
 		}
 	}
 
+	/*
+	 * If this message might trigger scheduler to give target (smtp),
+	 * we must tell the scheduler.
+	 */
+
+	for (h = e->e_headers; h != NULL; h = h->h_next) {
+		if (h->h_descriptor->hdr_name == NULL
+		    || !CISTREQ(h->h_descriptor->hdr_name, "turnme"))
+			continue;
+		for (ap = h->h_contents.a; ap != NULL; ap = ap->a_next) {
+			putc(_CF_TURNME, ofp);
+			putc(_CFTAG_NORMAL, ofp);
+			printAddress(ofp, ap->a_tokens, 0);
+			putc('\n', ofp);
+		}
+	}
+
 	FindEnvelope(eVerbose);
 	if (h != NULL
 	    && h->h_contents.a != NULL && h->h_contents.a->a_tokens != NULL) {
@@ -1903,3 +1924,33 @@ prdsndata(info, fp, cfflag, comment)
 	  if (*x->string != '\0')
 	    fprintf(fp, "%c %s\n", cfflag, x->string);
 }
+
+
+static struct conscell *
+find_errto(info)
+struct conscell *info;
+{
+	int i = 0;
+	register struct conscell *l, *x = NULL;
+
+	for (l = car(info); l != NULL; l = cdr(l)) {
+	  ++i;
+	  if (STRING(l)) {
+	    if (cdr(l) == NULL
+		&& (x = v_find(l->string)) != NULL
+		&& (x = cdr(x)) != NULL
+		&& LIST(x)) {
+	      for (x = car(x); x != NULL; x = cddr(x)) {
+		if (STRING(x) && strcmp((char *)x->string,"ERR") == 0) {
+		  x = cdr(x);
+		  break;
+		}
+	      }
+	      return x;
+	      /* if x == NULL, no Errors-To -expansion was specified */
+	    }
+	  } else
+	    return NULL; /* A list ??  Shall not be! */
+	}
+	return x;
+}
