@@ -273,7 +273,8 @@ mkMessageId(e, unixtime)
 {
 	register struct header *h;
 	static struct headerinfo *msgidDesc;
-	char *cp, *s, tzbuf[20];
+	static int genseq = 0;
+	char *cp, *s, *p, tzbuf[20];
 	u_char buf[1024];	/* XX: tsk tsk */
 	struct tm *ts;
 	
@@ -291,20 +292,22 @@ mkMessageId(e, unixtime)
 	h->h_next = 0;
 	ts = localtime(&unixtime);
 	cp = rfc822tz(ts, 1);
-	for (s = cp; *s != '\0'; ++s) {
+	for (s = cp, p = tzbuf; *s != '\0'; ++s) {
 		if (isascii(*s) && isupper(*s))
-			tzbuf[s-cp] = tolower(*s);
+			*p++ = tolower(*s);
+		else if (*s == '(' || *s == ')')
+			continue; /* Skip them.. */
 		else if (isascii(*s) && isspace(*s))
-			tzbuf[s-cp] = '_';
+			*p++ = '_';
 		else
-			tzbuf[s-cp] = *s;
+			*p++ = *s;
 	}
-	tzbuf[s-cp] = '\0';
+	*p = '\0';
 	cp = tzbuf;
-	sprintf((char *)buf, "<%02d%s%d.%02d%02d%02d%s.%s@%s>",
+	sprintf((char *)buf, "<%02d%s%d.%02d%02d%02d%s.%s+%d@%s>",
 		     ts->tm_year % 100, monthname[ts->tm_mon], ts->tm_mday,
 		     ts->tm_hour, ts->tm_min, ts->tm_sec,
-		     cp, e->e_file, myhostname);
+		     cp, e->e_file, ++genseq, myhostname);
 	h->h_lines = makeToken(buf, strlen((char *)buf));
 	h->h_lines->t_type = Line;
 	h->h_contents = hdr_scanparse(e, h, 0);
@@ -493,7 +496,15 @@ hdr_print(h, fp)
 			fputs(cp, fp);
 			putc('\n', fp);
 			break;
-		} /* else fall through */
+		} /* else we have the original line ... */
+		for (t = h->h_lines; t != NULL; t = t->t_next) {
+			cp = (char *)t->t_pname;
+			/* Skip over the leading white-space */
+			while (*cp == ' '||*cp == '\t') ++cp;
+			fputs(cp, fp);
+			putc('\n', fp);
+		}
+		break;
 	default:
 		for (t = h->h_lines; t != NULL; t = t->t_next) {
 			fwrite((char *)t->t_pname, sizeof (char),
@@ -755,7 +766,7 @@ saveAddress(pp)
 					*cp++ = '[';
 				else if (t->t_type == String)
 					*cp++ = '"';
-				bcopy((char *)t->t_pname, cp, len);
+				memcpy(cp, (char *)t->t_pname, len);
 				cp += len;
 				if (t->t_type == DomainLiteral)
 					*cp++ = ']';
