@@ -254,7 +254,7 @@ main(argc, argv)
 	      exit(EX_UNAVAILABLE);
 	    }
 	    sad.sin_port = port;
-	    bcopy(addr, (char *)&sad.sin_addr, hp->h_length);
+	    memcpy((char *)&sad.sin_addr, addr, hp->h_length);
 	    if (connect(fd, (char *)&sad, sizeof sad) < 0) {
 	      eval = errno;
 	      fprintf(stderr,
@@ -262,8 +262,7 @@ main(argc, argv)
 		      progname,
 		      dottedquad(&sad.sin_addr));
 	      if ((addr = *hp_nextaddr()) != NULL) {
-		bcopy(addr, (char *)&sad.sin_addr,
-		      hp->h_length);
+		memcpy((char *)&sad.sin_addr, addr, hp->h_length);
 		fprintf(stderr,
 			", trying %s...\n",
 			dottedquad(&sad.sin_addr));
