@@ -52,6 +52,7 @@ int	dumpq;
 int	canexit;
 int	rereadcf;
 int	verbose;
+int	dlyverbose = 0;
 int	querysocket = -1;	/* fd of TCP socket to listen for queries */
 int	subsubdirs;	/* use subdirectories under each channel directory */
 int	D_alloc = 0;
@@ -97,11 +98,11 @@ extern void deletemsg __((char *, struct ctlfile *));
 extern void detach __((void));
 extern void die __((int rc, char *reason));
 extern void doagenda __((void));
-extern void ipcpause __((void));
+extern void ipcpause __((int pollflag));
 extern void killprevious __((int signal, char *pidfile));
 extern void link_in __((int flag, struct vertex *vp, u_char *s));
 extern void mklink __((char *from, char *to));
-extern void mux __((void));
+extern void mux __((int quickcheck));
 extern void prversion __((char *name));
 extern void qprint __((int fd));
 extern void queryipcinit __((void));
@@ -133,10 +134,13 @@ main(argc, argv)
 
 	stickymem = MEM_MALLOC;
 
+	resources_maximize_nofiles();
+
 	postoffice = rendezvous = log = config = NULL;
 	daemonflg = 1;
+	dlyverbose = 0;
 	verbose = subsubdirs = errflg = version = 0;
-	while ((c = getopt(argc, argv, "disvf:L:P:Q:V")) != EOF)
+	while ((c = getopt(argc, argv, "disvf:L:P:Q:VW")) != EOF)
 		switch (c) {
 		case 'f':	/* override default config file */
 			config = optarg;
@@ -151,9 +155,13 @@ main(argc, argv)
 			rendezvous = optarg;
 			break;
 		case 'v':	/* be verbose and synchronous */
-			verbose = 1;
+			++verbose;
 			daemonflg = 0;
 			break;
+		case 'W':
+			dlyverbose = verbose;
+			verbose = 0;
+			break;
 		case 'd':	/* daemon again */
 			daemonflg = 1;
 			break;
@@ -277,12 +285,13 @@ main(argc, argv)
 	signal(SIGUSR2, sig_iot);	/* dump queue info */
 	cp = emalloc(strlen(TRANSPORTDIR)+4);
 	sprintf(cp, "../%s", TRANSPORTDIR);
+	queryipcinit();
 	syncweb(cp, 0, tp);
+	if (dlyverbose) verbose = dlyverbose;
 	free(cp);
 	canexit = 1;
 	signal(SIGTERM, sig_exit);	/* split */
 	sweepretry = 1;
-	queryipcinit();
 	alarm(sweepinterval);
 #ifdef	MALLOC_TRACE
 	mal_leaktrace(1);
@@ -305,8 +314,8 @@ main(argc, argv)
 				tp = rereadconfig(tp, config);
 				rereadcf = 0;
 			}
-			if (numkids > 0)
-				mux();
+			/* if (numkids > 0) */
+			mux(0);
 		}
 		canexit = 1; /* may be possible to move this after doagenda() */
 		/* wait for an alarm call */
@@ -319,7 +328,8 @@ main(argc, argv)
 				tp = rereadconfig(tp, config);
 				rereadcf = 0;
 			}
-			ipcpause();
+			/* ipcpause(0); */
+			mux(0);
 		}
 	}
 #ifdef	MALLOC_TRACE
@@ -392,6 +402,10 @@ syncweb(dir, nukefile, cehead)
 	for (dp = readdir(dirp); dp != NULL; dp = readdir(dirp)) {
 	  if (mustexit)
 	    break;
+
+	  /* ipcpause(1); */
+	  mux(1);
+
 	  if (isascii(dp->d_name[0]) && isdigit(dp->d_name[0])) {
 	    if (strcmp(dir, ".") == 0)
 	      strcpy(file, dp->d_name);
@@ -520,7 +534,7 @@ slurp(fd)
 	contents[stbuf.st_size] = 0;
 
 	cfp = (struct ctlfile *)emalloc((u_int) (sizeof(struct ctlfile)));
-	bzero((void*)cfp,(u_int) (sizeof(struct ctlfile)));
+	memset((void*)cfp, 0, sizeof(struct ctlfile));
 
 	cfp->fd = fd;
 	cfp->vfp = NULL;
@@ -624,7 +638,7 @@ vtxprep(cfp)
 	sp_install((u_int)cfp->id, (u_char *)cfp, 0, spt_mesh[L_CTLFILE]);
 	channel = host = NULL;
 	/* copy offsets into local array */
-	offspc = 20;
+	offspc = 16;
 	offarr = (struct offsort *)emalloc(offspc * sizeof(struct offsort));
 	opcnt = 0;
 	svn = 0;
@@ -652,7 +666,7 @@ vtxprep(cfp)
 	    switch (*--cp) {
 	    case _CF_RECIPIENT:
 	      if (opcnt >= offspc-1) {
-		offspc += 10;
+		offspc *= 2;
 		offarr = (struct offsort *)erealloc(offarr,
 						    sizeof(struct offsort) *
 						    offspc);
@@ -856,7 +870,7 @@ vtxprep(cfp)
 	      u_int alloc_size = (u_int) (sizeof (struct vertex) +
 					  (i - svn - 1) * sizeof (int));
 	      vp = (struct vertex *)emalloc(alloc_size);
-	      bzero((char*)vp,alloc_size);
+	      memset((char*)vp, 0, alloc_size);
 	      vp->cfp = cfp;
 	      defaultconfigentry(&vp->ce,NULL);
 	      vp->next[L_CTLFILE] = NULL;
@@ -898,7 +912,7 @@ vtxprep(cfp)
 	  u_int alloc_size = (u_int) (sizeof (struct vertex) +
 				      (i - svn - 1) * sizeof (int));
 	  vp = (struct vertex *)emalloc(alloc_size);
-	  bzero((void*)vp,alloc_size);
+	  memset((void*)vp, 0, alloc_size);
 	  vp->cfp = cfp;
 	  defaultconfigentry(&vp->ce,NULL);
 	  vp->next[L_CTLFILE] = NULL;
@@ -926,11 +940,11 @@ vtxprep(cfp)
 	}
 	*l_echannel = *l_ehost = ' ';
 	/*
-	  for (vp = head; vp != NULL; vp = vp->next[L_CTLFILE]) {
-	  printf("--\n");
-	  for (i = 0; i < vp->ngroup; ++i)
-	  printf("%s\n", cfp->contents+cfp->offset[vp->index[i]]);
-	}
+	   for (vp = head; vp != NULL; vp = vp->next[L_CTLFILE]) {
+	     printf("--\n");
+	     for (i = 0; i < vp->ngroup; ++i)
+	       printf("\t%s\n", cfp->contents+cfp->offset[vp->index[i]]);
+	   }
 	*/
 	cfp->head = head;
 	return cfp;
@@ -942,7 +956,7 @@ vtxfill(vp, tp)
 	struct config_entry *tp;
 {
 	/* if the channel doesn't match, there's no hope! */
-	if (verbose)
+	if (verbose>1)
 	  printf("ch? %s %s\n",
 		 vp->orig[L_CHANNEL]->name, tp->channel);
 	if (tp->channel[0] == '*' && tp->channel[1] == '\0')
@@ -956,7 +970,7 @@ vtxfill(vp, tp)
 	    return 0;
 	}
 
-	if (verbose)
+	if (verbose>1)
 	  printf("host %s %s\n", vp->orig[L_HOST]->name, tp->host);
 
 	if (tp->interval != -1) vp->ce.interval = tp->interval;
@@ -1037,7 +1051,7 @@ vtxdo(vp, cehead)
 	if (vp->ce.bychannel == -1)
 	  vp->ce.bychannel = 0;
 
-	if (verbose)
+	if (verbose>1)
 	  vtxprint(vp);
 
 	if ((vp->ce.flags & FF_GANGSCHEDULE)
@@ -1128,7 +1142,7 @@ link_in(flag, vp, s)
 	spl = sp_lookup(symbol(s), spt_mesh[flag]);
 	if (spl == NULL || (wp = (struct web *)spl->data) == NULL) {
 	  wp = (struct web *)emalloc(sizeof (struct web));
-	  bzero((void*)wp,sizeof (struct web));
+	  memset((void*)wp, 0, sizeof (struct web));
 	  sp_install(symbol(s), (u_char *)wp, 0, spt_mesh[flag]);
 	  wp->name = strsave((char *)s);
 	  wp->kids = 0;
@@ -1199,7 +1213,7 @@ mklink(from, to)
 	case ENOENT:	/* this may mean a subdirectory doesn't exist */
 	  for (cp = strchr(to, '/'); cp != NULL; cp = strchr(cp+1, '/')) {
 	    *cp = '\0';
-	    emkdir(to, 0755);
+	    mkdir(to, 0755);
 	    *cp = '/';
 	  }
 	  if (link(from, to) >= 0)
