@@ -20,6 +20,9 @@
 #define	FD_SETSIZE	32
 #endif	/* !FD_SETSIZE */
 
+#define TRANSPORTMAXNOFILES 10 /* Number of files a transporter may
+				  need open -- or any of its children.. */
+
 #ifdef	USE_UNIONWAIT
 #include <sys/wait.h>
 #endif	/* USE_UNIONWAIT */
@@ -27,12 +30,11 @@
 struct procinfo *cpids = NULL;
 
 
-#define	PIPESIZ	4096	/* pipe buffer size (max contents before blocking) */
-
 #define	MAXFILESPERTRANSPORT	1000
 
 int	numkids = 0;
 
+extern int resources_query_pipesize();
 extern char *getzenv();
 extern char *progname;
 extern char *replchannel;
@@ -215,6 +217,11 @@ runcommand(filearr, nelem, argv, vhead, chwp, howp)
 	int	i, pid, position, to[2], from[2], uid, gid;
 	char	*cmd;
 	FILE	*fp;
+	static int pipesize;
+
+	if (pipesize == 0)
+	  pipesize = resources_query_pipesize();
+
 
 	uid = vhead->ce.uid;
 	gid = vhead->ce.gid;
@@ -233,6 +240,7 @@ runcommand(filearr, nelem, argv, vhead, chwp, howp)
 	  fprintf(stderr, "to %d/%d from %d/%d\n",
 		  to[0],to[1],from[0],from[1]);
 	if ((pid = fork()) == 0) {	/* child */
+	  int topfileno;
 	  if (to[0] != 0) {
 	    dup2(to[0],0);	/* child stdin */
 	    close(to[0]);
@@ -247,13 +255,15 @@ runcommand(filearr, nelem, argv, vhead, chwp, howp)
 #if	1
 	  /* ... if detach() did its job, there shouldn't be any! */
 	  /* ... no, the 'querysock' is there somewhere!   */
-	  for (i = 3; i < getdtablesize(); ++i)
+	  topfileno = resources_query_nofiles();
+	  for (i = 3; i < topfileno; ++i)
 	    close(i);
 #else
 	  close(to[1]);
 	  close(from[0]);
 	  /* we can ignore the stdouts from other children */
 #endif
+	  resources_limit_nofiles(TRANSPORTMAXNOFILES);
 	  setgid(gid);	/* Do GID setup while still UID 0..   */
 	  setuid(uid);	/* Now discard all excessive powers.. */
 	  execv(cmd, argv);
@@ -283,7 +293,7 @@ runcommand(filearr, nelem, argv, vhead, chwp, howp)
 	  }
 	  for (i = position = 0; i < nelem; ++i) {
 	    position += strlen(filearr[i]) + 1;
-	    if (position + 1 >= PIPESIZ)
+	    if (position + 1 >= pipesize)
 	      break;
 	    fprintf(fp, "%s\n", filearr[i]);
 	  }
@@ -350,15 +360,19 @@ reclaim(fd)
 	  --cpids[fd].ch->kids;
 	  if (cpids[fd].ch->fifohead)
 	    fifoschedule(cpids[fd].ch);
-	  else if (cpids[fd].ch->kids == 0 && cpids[fd].ch->link == NULL)
+	  else if (cpids[fd].ch->kids == 0 && cpids[fd].ch->link == NULL) {
 	    unweb(L_CHANNEL, cpids[fd].ch);
+	    cpids[fd].ch = NULL;
+	  }
 	}
 	if (cpids[fd].ho != NULL) {
 	  --cpids[fd].ho->kids;
 	  if (cpids[fd].ho->fifohead)
 	    fifoschedule(cpids[fd].ho);
-	  else if (cpids[fd].ho->kids == 0 && cpids[fd].ho->link == NULL)
+	  else if (cpids[fd].ho->kids == 0 && cpids[fd].ho->link == NULL) {
 	    unweb(L_HOST, cpids[fd].ho);
+	    cpids[fd].ho = NULL;
+	  }
 	}
 	close(fd);
 
@@ -462,20 +476,28 @@ typedef	struct fd_set { fd_mask	fds_bits[1]; } fd_set;
 #endif	/* !NFDBITS */
 
 void
-mux()
+mux(quickcheck)
+int quickcheck;
 {
 	int	i, n, rc, maxf;
 	fd_set	mask;
 	struct sockaddr_in raddr;
 	int	raddrlen;
+	struct timeval tv;
+
+	tv.tv_sec = 10000; /* LONG delay.. */
+	tv.tv_usec = 0;
+	if (quickcheck)
+	  tv.tv_sec = 0; /* Check the sockets fast.. */
 
 	maxf = -1;
 	FD_ZERO(&mask);
-	for (i = 0; i < FD_SETSIZE ; ++i)
-	  if (cpids[i].pid != 0) {
-	    FD_SET(i, &mask);
-	    maxf = i;
-	  }
+	if (cpids != NULL)
+	  for (i = 0; i < FD_SETSIZE ; ++i)
+	    if (cpids[i].pid != 0) {
+	      FD_SET(i, &mask);
+	      maxf = i;
+	    }
 	if (querysocket >= 0) {
 	  FD_SET(querysocket, &mask);
 	  if (maxf < querysocket)
@@ -486,87 +508,93 @@ mux()
 	++maxf;
 	/*fprintf(stderr, "about to select on %x [%d]\n",
 			  mask.fds_bits[0], maxf); */
-	if ((n = select(maxf, &mask, NULL, NULL, NULL)) < 0) {
+	if ((n = select(maxf, &mask, NULL, NULL, &tv)) < 0) {
 	  /* fprintf(stderr, "got an interrupt (%d)\n", errno); */
 	  if (errno == EINTR)
 	    return;
 	} else if (n == 0) {
-	  fprintf(stderr, "abnormal 0 return from select!\n");
+	  /* fprintf(stderr, "abnormal 0 return from select!\n"); */
+	  /* -- just a timeout -- fast or long */
+	  return;
 	} else {
 	  /*fprintf(stderr, "got %d ready (%x)\n", n, mask.fds_bits[0]);*/
-	  while (n > 0) {
-	    if (querysocket >= 0 && FD_ISSET(querysocket, &mask)) {
-	      --n;
-	      FD_CLR(querysocket, &mask);
-	      raddrlen = sizeof raddr;
-	      i = accept(querysocket,
-			 (struct sockaddr *)&raddr,&raddrlen);
-	      if (i < 0) {
-		perror("accept");
-	      } else {
-		/* rc = fork();
-		   if (rc == 0) { /* Child! */
+	  if (querysocket >= 0 && FD_ISSET(querysocket, &mask)) {
+	    --n;
+	    FD_CLR(querysocket, &mask);
+	    raddrlen = sizeof raddr;
+	    i = accept(querysocket,
+		       (struct sockaddr *)&raddr,&raddrlen);
+	    if (i < 0) {
+	      perror("accept");
+	    } else {
+	      rc = fork();
+	      if (rc == 0) { /* Child! */
 		qprint(i);
-		/* exit(0);
-		   } 
-		   close(i); */
+		exit(0);
 	      }
-	      continue;
+	      close(i);
 	    }
+	  }
+	  if (cpids != NULL)
 	    for (i = 0; i < maxf; ++i) {
 	      if (cpids[i].pid != 0 && FD_ISSET(i, &mask)) {
 		--n;
 		FD_CLR(i, &mask);
 		/*fprintf(stderr,"that is fd %d\n",i);*/
-		break;
+		/* do non-blocking reads from this fd */
+		readfrom(i);
 	      }
 	    }
-	    if (i >= maxf)
-	      break;
-	    /* do non-blocking reads from this fd */
-	    readfrom(i);
-	  }
 	}
 	/* fprintf(stderr, "return from mux\n"); */
 }
 
 void
-ipcpause()
+ipcpause(pollflag)
+int pollflag;
 {
 #ifdef	AF_INET
 	int	i, n, rc;
 	fd_set	mask;
 	struct sockaddr_in raddr;
 	int	raddrlen;
+	struct timeval timeout;
 
-	if (querysocket < 0) {
+	if (querysocket < 0 && !pollflag) {
 		pause();
 		return;
 	}
 	FD_ZERO(&mask);
 	FD_SET(querysocket, &mask);
-	if ((n = select(querysocket+1, &mask, NULL, NULL, NULL)) < 0) {
+	timeout.tv_sec = 0;
+	timeout.tv_usec = 0;
+	if (!pollflag)
+	  timeout.tv_sec = 9999; /* VERY long time..
+				    We wait for an interrupt.. */
+	if ((n = select(querysocket+1, &mask, NULL, NULL, &timeout)) < 0) {
 	  if (errno == EINTR)
 	    return;
 	} else if (n == 0) {
-	  fprintf(stderr, "abnormal 0 return from select!\n");
+	  /* fprintf(stderr, "abnormal 0 return from select!\n"); */
+	  return;
 	} else {
 	  raddrlen = sizeof raddr;
 	  i = accept(querysocket, (struct sockaddr *)&raddr, &raddrlen);
 	  if (i < 0) {
-	    perror("accept");
+	    perror("querysocket accept");
 	  } else {
 	    /* Fork for QPRINT */
-	    /* rc = fork();
-	       if (rc == 0) { /* Child! */
-	    qprint(i);
-	    /* exit(0);
-	       }
-	       close(i); */
+	    rc = fork();
+	    if (rc == 0) { /* Child! */
+	      qprint(i);
+	      exit(0);
+	    }
+	    close(i);
 	  }
 	}
 #else	/* !AF_INET */
-	pause();
+	if (!pollflag)
+	  pause();
 #endif	/* AF_INET */
 }
 
@@ -612,7 +640,8 @@ queryipcinit()
 }
 #else	/* !HAVE_SELECT */
 void
-mux()
+mux(quickcheck)
+int quickcheck; /* Is always a quickcheck..  Aurgh the induced load.. */
 {
 	int	fd;
 
@@ -623,17 +652,17 @@ mux()
 	 * are possible in the absence of async i/o like select() or the
 	 * simulation USG supplies, but it ain't worth the hassle.
 	 */
-	for (fd = 0; fd < FD_SETSIZE ; ++fd)
-	  if (cpids[fd].pid != 0)
-	    break;
-	if (fd >= FD_SETSIZE)
-	  return;
-	readfrom(fd);
+	if (cpids != NULL)
+	  for (fd = 0; fd < FD_SETSIZE ; ++fd)
+	    if (cpids[fd].pid != 0)
+	      readfrom(fd);
 }
 
-ipcpause()
+ipcpause(pollflag)
+int pollflag;
 {
-	pause();
+	if (!pollflag)
+		pause();
 }
 
 void
@@ -699,7 +728,7 @@ readfrom(fd)
 	      pcp = buf = erealloc(buf,bufsize);
 	      cp = buf + (bufsize - 1024);
 	    } else {
-	      bcopy(pcp, buf, cp-pcp);
+	      memcpy(buf, pcp, cp-pcp);
 	      cp = buf + (cp-pcp);
 	      *cp = '\0';
 	      pcp = buf;	/* may be used below */
@@ -728,7 +757,7 @@ readfrom(fd)
 	if (dontbreak) {
 	  if (cpids[fd].pid != 0) {
 	    cpids[fd].carryover = emalloc(cp-pcp+1);
-	    bcopy(pcp, cpids[fd].carryover, cp-pcp);
+	    memcpy(cpids[fd].carryover, pcp, cp-pcp);
 	    *(cpids[fd].carryover+(cp-pcp)) = '\0';
 	  } else
 	    fprintf(stderr,
