@@ -213,6 +213,8 @@ char buf[SMTPLINESIZE];	/* limits size of SMTP commands...
 			   On the other hand, limit is asked
 			   to be only 1000 chars, not 8k.. */
 char helobuf[SMTPLINESIZE];
+int maxloadavg = 10;	/* Maximum load-average that is tolerated
+			   with smtp-server actively receiving.. */
 
 
 
@@ -266,6 +268,7 @@ extern char *strchr();
 #define	putc	fputc
 #endif	/* lint */
 
+char *msg_toohighload = "400 Sorry, system is too loaded for email reception at the moment\r\n"; /* XX: ??? */
 
 
 void
@@ -318,7 +321,7 @@ main(argc, argv)
 				       progname, myhostname, strerror(errno));
 		exit(1);
 	}
-	while ((c = getopt(argc, argv, "aBigl:np:P:R:s:Vv")) != EOF) {
+	while ((c = getopt(argc, argv, "aBigl:np:L:P:R:s:Vv")) != EOF) {
 		switch (c) {
 #ifdef	USE_AUTH
 		case 'a':
@@ -348,6 +351,11 @@ main(argc, argv)
 			style = emalloc(strlen(optarg)+1);
 			strcpy(style, optarg);
 			break;
+		case 'L':	/* Max LoadAverage */
+			maxloadavg = atoi(optarg);
+			if (maxloadavg < 1)
+			  maxloadavg = 10; /* Humph.. */
+			break;
 #ifdef	USE_INET
 		case 'p':
 			port = htons((u_short)atoi(optarg));
@@ -385,11 +393,12 @@ main(argc, argv)
 	if (errflg || optind != argc) {
 #ifdef	USE_INET
 		fprintf(stderr,
-			       "Usage: %s [-ign] [-p port#] [-l logfile]\n",
-			       progname);
+			"Usage: %s [-aBivgnV] [-s xx] [-L maxLoadAvg] [-R rtrprog] [-p port#] [-P postoffice] [-l logfile]\n",
+			progname);
 #else	/* !USE_INET */
 		fprintf(stderr,
-			       "Usage: %s [-ign] [-l logfile]\n", progname);
+			"Usage: %s [-aBivgnV] [-s xx] [-L maxLoadAvg] [-R rtrprog] [-P postoffice] [-l logfile]\n",
+			progname);
 #endif	/* USE_INET */
 		exit(1);
 	}
@@ -418,10 +427,15 @@ main(argc, argv)
 	} else
 #ifdef	USE_INET
 	if (inetd) {
+		if (maxloadavg < loadavg_current()) {
+		  write(1, msg_toohighload, strlen(msg_toohighload));
+		  exit(1);
+		}
+
 		raddrlen = sizeof raddr;
 		if (getpeername(0,(struct sockaddr *)&raddr, &raddrlen) == -1) {
 			fprintf(stderr, "%s: getpeername(0): %s\n",
-					       progname, strerror(errno));
+				progname, strerror(errno));
 			exit(1);
 		}
 		setrhostname(&raddr);
@@ -429,8 +443,7 @@ main(argc, argv)
 		if (auth_flag != 0) {
 #ifdef  LOG_PID
 #ifdef  LOG_MAIL
-			openlog("smtpserver", LOG_PID|LOG_CONS,
-				LOG_MAIL);
+			openlog("smtpserver", LOG_PID|LOG_CONS, LOG_MAIL);
 #else  /* !LOG_MAIL */
 			openlog("smtpserver", 0);
 #endif /* LOG_MAIL */
@@ -461,15 +474,15 @@ main(argc, argv)
 		killprevious(SIGTERM, PID_SMTPSERVER);
 		if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
 			fprintf(stderr,
-				       "%s: socket(AF_INET, SOCK_STREAM): %s\n",
-				       progname, strerror(errno));
+				"%s: socket(AF_INET, SOCK_STREAM): %s\n",
+				progname, strerror(errno));
 			exit(1);
 		}
 		i = 1;
 		if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (caddr_t)&i, sizeof i) < 0) {
 			fprintf(stderr,
-				       "%s: setsockopt(SO_REUSEADDR): %s\n",
-				       progname, strerror(errno));
+				"%s: setsockopt(SO_REUSEADDR): %s\n",
+				progname, strerror(errno));
 			exit(1);
 		}
 		if (port <= 0) {
@@ -480,7 +493,8 @@ main(argc, argv)
 
 			if ((service = getservbyname("smtp", "tcp")) == NULL) {
 				fprintf(stderr,
-				    "%s: no SMTP service entry\n", progname);
+					"%s: no SMTP service entry\n",
+					progname);
 				exit(1);
 			}
 			port = service->s_port;
@@ -491,12 +505,12 @@ main(argc, argv)
 		sad.sin_port = port;
 		if (bind(s, (struct sockaddr *)&sad, sizeof sad) < 0) {
 			fprintf(stderr, "%s: bind(): %s\n",
-					       progname, strerror(errno));
+				progname, strerror(errno));
 			exit(1);
 		}
 		if (listen(s, 5) < 0) {
 			fprintf(stderr, "%s: listen(): %s\n",
-					       progname, strerror(errno));
+				progname, strerror(errno));
 			exit(1);
 		}
 		settrusteduser();	/* dig out the trusted user ID */
@@ -511,9 +525,13 @@ main(argc, argv)
 					fflush(stderr);
 				close(msgfd);
 			}
+#if 0
+		signal(SIGCHLD, SIG_IGN);
+#else
 		signal(SIGCHLD, reaper);
+#endif
 		signal(SIGALRM, timedout);
-		signal(SIGHUP, SIG_IGN);
+		signal(SIGHUP,  SIG_IGN);
 		signal(SIGTERM, SIG_DFL);
 		while (1) {
 			int childpid;
@@ -524,21 +542,28 @@ main(argc, argv)
 					continue;
 				else {
 					fprintf(stderr,
-						       "%s: accept(): %s\n",
-						       progname,
-						       strerror(errno));
+						"%s: accept(): %s\n",
+						progname,
+						strerror(errno));
 					exit(1);
 				}
 			}
 			if ((childpid = fork()) < 0) {	/* can't fork! */
 				close(msgfd);
 				fprintf(stderr,
-					       "%s: fork(): %s\n",
-					       progname, strerror(errno));
+					"%s: fork(): %s\n",
+					progname, strerror(errno));
 				sleep(5);
 				continue;
 			} else if (childpid == 0) {		/* child */
 				close(s);
+
+				if (maxloadavg < loadavg_current()) {
+				  write(msgfd, msg_toohighload,
+					strlen(msg_toohighload));
+				  exit(1);
+				}
+
 				signal(SIGTERM, SIG_IGN);
 				if (msgfd != 0)
 					dup2(msgfd, 0);
@@ -1296,6 +1321,11 @@ smtpserver(insecure)
 					 eobuf[-1] == '\t'))
 		  *--eobuf = '\0';
 
+		if (logfp != NULL) {
+			fprintf(logfp, "%dr\t%s\n", pid, buf);
+			fflush(logfp);
+		}
+
 		if (rc >= 0) {
 		  rfc821_error_ptr = buf + rc;
 		  type821err(500, buf,
@@ -1310,10 +1340,6 @@ smtpserver(insecure)
 			fprintf(stdout,"%s\n",buf); /* XX: trace.. */
 		report("%.100s", buf);
 
-		if (logfp != NULL) {
-			fprintf(logfp, "%dr\t%s\n", pid, buf);
-			fflush(logfp);
-		}
 		for (cp = buf; isascii(*cp) && isalpha(*cp); ++cp)
 			continue;
 		if (cp >  buf + 6) { /* "DEBUG" is longest of them.. */
@@ -1782,7 +1808,11 @@ callr()
 	if (pipe(to) < 0 || pipe(from) < 0)
 		return -1;
 
+#if 0
+	signal(SIGCHLD, SIG_IGN);
+#else
 	signal(SIGCHLD, reaper);
+#endif
 	signal(SIGPIPE, SIG_IGN);
 
 	if (routerprog == NULL) {
