@@ -1,6 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Copyright 1992-1994 Matti Aarnio -- MIME processing et.al.
  */
 
 /* History:
@@ -207,6 +208,7 @@ int creatflag = 1;		/* attempt to create files that don't exist */
 int mime8bit = 0;		/* Has code which translates  MIME text/plain
 				   Quoted-Printable  to 8BIT.. */
 int convert_qp = 0;		/* Flag: We have a job.. */
+int conversion_prohibited = 0;	/* Flag: Under no circumstances touch on it.. */
 int is_mime    = 0;		/* Flag: Msg is MIME msg .. */
 char *mime_boundary = NULL;
 int   mime_boundary_len = 0;
@@ -218,18 +220,22 @@ extern time_t time();
 extern int optind;
 extern char *optarg;
 extern int emptyline();
-extern void prversion(), getnobody(), process(), biff(), rbiff();
-extern void deliver();
+extern void prversion(), getnobody(), biff(), rbiff();
 extern struct passwd *getpwnam();
-extern int setupuidgid(), createfile();
-extern int exstat(), creatembox();
-extern char *exists(), *strerror();
-extern FILE *putmail();
-extern void program(), setrootuid();
-extern int appendlet();
-extern char **environ;
+extern int setupuidgid();
+extern int createfile();
 extern int exstat();
-extern int writebuf();
+extern int creatembox();
+extern char *exists(), *strerror();
+extern void setrootuid();
+extern void process __((struct ctldesc *dp));
+extern void deliver __((struct ctldesc *dp, struct rcpt *rp, char *timestring));
+extern void program __((struct ctldesc *dp, struct rcpt *rp, char *timestring));
+extern FILE *putmail __((struct ctldesc *dp, struct rcpt *rp, int fdmail, char *timestring, char *file));
+extern int appendlet __((struct ctldesc *dp, struct rcpt *rp, FILE *fp, char *file, int ismime));
+extern char **environ;
+extern int writebuf __((FILE *fp, char *buf, int len));
+extern int writemimebuf __((FILE *fp, char *buf, int len));
 
 extern int  qp_to_8bit   __((struct rcpt *rp));
 extern int  qptext_check __((struct rcpt *rp));
@@ -405,9 +411,11 @@ process(dp)
 	char *cp, *ts, *at;
 	struct rcpt *rp;
 	time_t curtime;
-	int is_mime_qptext;
+	int is_mime_qptext = 0;
 
-	is_mime_qptext = qptext_check(dp->recipients);
+	conversion_prohibited = check_conv_prohibit(dp->recipients);
+	if (!conversion_prohibited)
+	  is_mime_qptext = qptext_check(dp->recipients);
 
 	/*
 	 * Strip backslashes prefixed to the user name,
@@ -450,7 +458,7 @@ process(dp)
 	      warning("Null message offset in \"%s\"!",
 		      dp->msgfile);
 	    convert_qp = is_mime_qptext && mime8bit;
-	    deliver(rp, dp, ts);
+	    deliver(dp, rp, ts);
 	  }
 	}
 }
@@ -461,9 +469,9 @@ process(dp)
  */
 
 void
-deliver(rp, dp, timestring)
-	struct rcpt *rp;
+deliver(dp, rp, timestring)
 	struct ctldesc *dp;
+	struct rcpt *rp;
 	char *timestring;
 {
 	register char **maild;
@@ -494,7 +502,7 @@ deliver(rp, dp, timestring)
 		       "mail to program disallowed", 0);
 	    return;
 	  }
-	  program(rp, messagefd, timestring, dp);
+	  program(dp, rp, timestring);
 	  /* DIAGNOSTIC(rp, EX_UNAVAILABLE,
 	     "mailing to programs (%s) is not supported",
 	     rp->addr->user+1); */
@@ -756,7 +764,7 @@ deliver(rp, dp, timestring)
 	if (nbp != NULL)
 	  nbp->offset = lseek(fdmail, 0L, L_XTND);
 #endif	/* BIFF || RBIFF */
-	fp = putmail(rp, fdmail, messagefd, timestring, file, dp);
+	fp = putmail(dp, rp, fdmail, timestring, file);
 	if (mmdf_mode)
 	  fputs("\001\001\001\001\n",fp);
 	
@@ -793,23 +801,18 @@ deliver(rp, dp, timestring)
 }
 
 FILE *
-putmail(rp, fdmail, messagefd, timestring, file, dp)
-	struct rcpt *rp;
-	int fdmail, messagefd;
-	char *timestring, *file;
+putmail(dp, rp, fdmail, timestring, file)
 	struct ctldesc *dp;
+	struct rcpt *rp;
+	int fdmail;
+	char *timestring, *file;
 {
 	FILE *fp;
-	int len;
+	int len, rc;
 	long eofindex;
 	char buf[2];
 	struct stat st;
 	char *fromuser;
-	time_t orig_atime = 0;
-
-	if (keepatime  &&  *(rp->addr->user) != TO_PIPE  &&
-	    stat(file, &st) >= 0)
-	  orig_atime = st.st_atime;
 
 	if ((fp = fdopen(fdmail, "a+")) == NULL) {
 	  notaryreport(NULL,NULL,NULL);
@@ -833,27 +836,31 @@ putmail(rp, fdmail, messagefd, timestring, file, dp)
 	   * login etc. can distinguish new mail from old.
 	   * The mtime will be set to now by the following write() calls.
 	   */
+
 	  fstat(fileno(fp), &st);
+	  if (!keepatime) st.st_atime = 0;
 
 	  len = fread(buf, 1, 2, fp);
-	  if (!keepatime) {
+	  fseek(fp, 0L, L_XTND);	/* to end of file, again */
+
+	  if (st.st_atime != 0) {
 #ifdef	USE_TIMEVAL
 	    struct timeval tv[2];
-	    tv[0].tv_sec = orig_atime;
+	    tv[0].tv_sec = st.st_atime;
 	    tv[1].tv_sec = st.st_mtime;
 	    tv[0].tv_usec = tv[1].tv_usec = 0;
-	    utimes(file, tv);
+	    rc = utimes(file, tv);
 #else  /* !USE_TIMEVAL */
 	    struct utimbuf tv;
-	    tv.actime  = orig_atime;
+	    tv.actime  = st.st_atime;
 	    tv.modtime = st.st_mtime;
-	    utime(file, &tv);
+	    rc = utime(file, &tv);
 #endif /* USE_TIMEVAL */
 	  }
+
 	  if (len == 1 || len == 2) {
 	    --len;
 	    len = (buf[len]!='\n') + (len == 1 ? buf[0]!='\n' : 1);
-	    fseek(fp, 0L, L_XTND);	/* to end of file, again */
 	    if (len > 0 && (fwrite("\n\n", 1, len, fp) != len
 			    || fflush(fp) == EOF)) {
 
@@ -868,14 +875,14 @@ putmail(rp, fdmail, messagefd, timestring, file, dp)
 	      fclose(fp);
 	      return NULL;
 	    }
-	  } else
-	    fseek(fp, 0L, L_XTND);	/* to end of file, again */
+	  }
 	} else if (eofindex > 0L && eofindex < (sizeof "From x\n")) {
 	  /* A mail message *cannot* be this small.  It must be
 	     a corrupted mailbox file.  Ignore the trash bytes. */
 	  fseek(fp, 0L, L_SET);
 	  eofindex = 0;
 	}
+
 	if (!(convert_qp && qp_to_8bit(rp)))
 	  convert_qp = 0; /* If malloc failed, clear this too,
 			     if it was not set, it doesn't change.. */
@@ -900,7 +907,7 @@ putmail(rp, fdmail, messagefd, timestring, file, dp)
 	  fclose(fp);
 	  return NULL;
 	}
-	if (appendlet(rp, fp, file, messagefd, is_mime) == EOF) {
+	if (appendlet(dp, rp, fp, file, is_mime) == EOF) {
 	  /* XX: should I really do this? */
 	  fflush(fp);
 	  ftruncate(fileno(fp), (u_long)eofindex);
@@ -933,15 +940,15 @@ putmail(rp, fdmail, messagefd, timestring, file, dp)
 #endif
 	  fflush(logfp);
 	}
+
 	return fp;
 }
 
 void
-program(rp, messagefd, timestring, dp)
-	struct rcpt *rp;
-	int messagefd;
-	char *timestring;
+program(dp, rp, timestring)
 	struct ctldesc *dp;
+	struct rcpt *rp;
+	char *timestring;
 {
 	int i, pid, uid, gid, in[2], out[2];
 	char *env[20], buf[8192], *cp, *s;
@@ -1057,7 +1064,7 @@ program(rp, messagefd, timestring, dp)
 	errfp = fdopen(in[0], "r");
 	/* write the message */
 	mmdf_mode += 2;
-	fp = putmail(rp, out[1], messagefd, timestring, rp->addr->user+1, dp);
+	fp = putmail(dp, rp, out[1], timestring, rp->addr->user+1);
 	mmdf_mode -= 2;
 	if (fp == NULL) {
 	  pid = wait(&status);
@@ -1149,14 +1156,14 @@ creatembox(rp, filep, uid, gid, pw)
 #ifdef	USE_TIMEVAL
 	      struct timeval tv[2];
 	      stat(*filep,&st); /* This by all propability will not fail.. */
-	      tv[0].tv_sec = 0;
+	      tv[0].tv_sec = 1; /* Not zero, but close.. */
 	      tv[1].tv_sec = st.st_mtime;
 	      tv[0].tv_usec = tv[1].tv_usec = 0;
 	      utimes(*filep, tv);
 #else
 	      struct utimbuf tv;
 	      stat(*filep,&st); /* This by all propability will not fail.. */
-	      tv.actime  = 0;
+	      tv.actime  = 1; /* Not zero, but close.. */
 	      tv.modtime = st.st_mtime;
 	      utime(*filep, &tv);
 #endif
@@ -1309,12 +1316,14 @@ setrootuid(rp)
  * appendlet - append letter to file pointed at by fd
  */
 int
-appendlet(rp, fp, file, mfd, ismime)
+appendlet(dp, rp, fp, file, ismime)
+	struct ctldesc *dp;
 	struct rcpt *rp;
 	FILE *fp;
-	int mfd, ismime;
+	int ismime;
 	char *file;
 {
+#ifndef	USE_MMAP
 	static char buffer[BUFSIZ+1];
 	static char *bufend = buffer+BUFSIZ;
 	register int i;
@@ -1322,10 +1331,13 @@ appendlet(rp, fp, file, mfd, ismime)
 	char iobuf[BUFSIZ];
 	FILE *mfp;
 
-	writebuf((FILE *)NULL, (char *)NULL, 0, 0);	/* initialize state */
+	if (ismime)
+	  writemimebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
+	else
+	  writebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
 
 	if (ismime) {
-	  /* can we use cache of message body data */
+	  /* can we use cache of message body data ? */
 	  /* Split it to lines.. */
 	  if (readalready > 0) {
 	    int linelen;
@@ -1339,7 +1351,7 @@ appendlet(rp, fp, file, mfd, ismime)
 	      if (*s == '\n' && readalready > 0) {
 		++s; ++linelen; --readalready;
 	      }
-	      if (writebuf(fp, s0, linelen, ismime) != linelen) {
+	      if (writemimebuf(fp, s0, linelen) != linelen) {
 		DIAGNOSTIC(rp, EX_IOERR,
 			   "write to \"%s\" failed", file);
 		return EOF;
@@ -1366,10 +1378,10 @@ appendlet(rp, fp, file, mfd, ismime)
 	    if (i == BUFSIZ && buffer[BUFSIZ-1] != '\n')
 	      ismime = 0;
 	    /* Ok, write the line */
-	    if (writebuf(fp, buffer, i, ismime) != i) {
+	    if (writemimebuf(fp, buffer, i) != i) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
-	      MFPCLOSE;
+ 	      MFPCLOSE;
 	      return EOF;
 	    }
 	  }
@@ -1386,7 +1398,7 @@ appendlet(rp, fp, file, mfd, ismime)
 
 	  /* can we use cache of message body data */
 	  if (readalready != 0) {
-	    if (writebuf(fp, buffer, readalready, 0) != readalready) {
+	    if (writebuf(fp, buffer, readalready) != readalready) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
 	      return EOF;
@@ -1404,7 +1416,7 @@ appendlet(rp, fp, file, mfd, ismime)
 			 "read error from message file", 0);
 	      return EOF;
 	    }
-	    if (writebuf(fp, buffer, i, 0) != i) {
+	    if (writebuf(fp, buffer, i) != i) {
 	      DIAGNOSTIC(rp, EX_IOERR,
 			 "write to \"%s\" failed", file);
 	      return EOF;
@@ -1416,6 +1428,33 @@ appendlet(rp, fp, file, mfd, ismime)
 	  if (bufferfull > 1)	/* not all in memory, need to reread */
 	    readalready = 0;
 	}
+#else /* USE_MMAP  --  get the input from  mmap():ed memory area.. */
+	char *s;
+	s = dp->let_buffer + dp->msgbodyoffset;
+	if (ismime) {
+	  int i;
+	  writemimebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
+	  while (s < dp->let_end) {
+	    char *s2 = s;
+	    i = 0;
+	    while (s2 < dp->let_end && *s2 != '\n') ++s2, ++i;
+	    if (s2 < dp->let_end && *s2 == '\n') ++s2, ++i;
+	    if (writemimebuf(fp, s, i) != i) {
+	      DIAGNOSTIC(rp, EX_IOERR,
+			 "write to \"%s\" failed", file);
+	      return EOF;
+	    }
+	    s = s2;
+	  }
+	} else {
+	  writebuf((FILE *)NULL, (char *)NULL, 0);	/* initialize state */
+	  if (writebuf(fp, s, dp->let_end - s) != (dp->let_end - s)) {
+	      DIAGNOSTIC(rp, EX_IOERR,
+			 "write to \"%s\" failed", file);
+	      return EOF;
+	  }
+	}
+#endif
 	return 0;
 }
 
@@ -1467,7 +1506,7 @@ biff(hostname, username, offset)
 	  bzero((char *)&biffaddr, sizeof biffaddr);
 	  biffaddr.sin_family = hp->h_addrtype;
 	  hp_init(hp);
-	  bcopy(*hp_getaddr(), (char *)&biffaddr.sin_addr, hp->h_length);
+	  memcpy((char *)&biffaddr.sin_addr, *hp_getaddr(),  hp->h_length);
 	  biffaddr.sin_port = sp->s_port;
 	  bap = (struct sockaddr_in *)emalloc(sizeof(struct sockaddr_in));
 	  *bap = biffaddr;
@@ -1575,7 +1614,7 @@ rbiff(nbp)
  */
 
 int
-writebuf(fp, buf, len, ismime)
+writebuf(fp, buf, len)
 	FILE *fp;
 	char *buf;
 	int len;
@@ -1612,73 +1651,8 @@ writebuf(fp, buf, len, ismime)
 	}
 #endif
 
-	if (ismime) {
-	  /* ------------------------------------------------------------ */
-	  /* MIME format processing, the input is LINE BY LINE */
-	  if (mime_boundary != NULL && n > mime_boundary_len+2 &&
-	      buf[0] == '-' && buf[1] == '-' &&
-	      strncmp(buf+2,mime_boundary,mime_boundary_len) == 0) {
-	    /* ------------------------------------------------------------ */
-	    /* XX: process MIME boundary! */
-	    i = fwrite(buf, 1, len, fp);
-	    if (i != len) return -1;
-	    return len;
-	  } else if (convert_qp) {
-	    /* ------------------------------------------------------------ */
-	    /* Process Q-P chars of this line */
-	    int qp_hex = 0;
-	    int qp_chrs = 0;
-	    int tlen = 0;
-
-	    for (cp = buf, s = buf2, n = len; n > 0; --n, ++cp) {
-	      register int c = *cp;
-	      /* At first, convert the possible QP character.. */
-	      if (!qp_chrs && c == '=') {
-		qp_chrs = 2;
-		qp_hex  = 0;
-		continue;	/* '=' starts something */
-	      }
-	      if (qp_chrs && c == '\n') {
-		qp_chrs = 0;	/* It was "=[ \t]*\n", which was thrown away.*/
-		continue;
-	      }
-	      if (qp_chrs == 2 && (c == ' ' || c == '\t'))
-		continue; /* It is "=[ \t]*\n", throw it away */
-	      /* Ok, this may be of  =HexHex ? */
-	      if (qp_chrs && ((c >= '0' && c <= '9') ||
-			      (c >= 'A' && c <= 'F') ||
-			      (c >= 'a' && c <= 'f'))) {
-		qp_hex <<= 4;
-		if (c >= '0' && c <= '9') qp_hex += (c - '0');
-		if (c >= 'A' && c <= 'F') qp_hex += (c - 'A' + 10);
-		if (c >= 'a' && c <= 'f') qp_hex += (c - 'a' + 10);
-		--qp_chrs;
-		if (!qp_chrs) c = qp_hex;
-		else continue;	/* Go collect another digit */
-	      } else if (qp_chrs)
-		qp_chrs = 0; /* Failed a HEX check.. */
-	      /* Now we have a converted character at `c', save it! */
-	      *s++ = c;
-	      ++tlen;
-	    }
-	    /* Uhh... Stored the conversion result into buf2[], length  tlen */
-	    if (!mmdf_mode && buf2[0] == 'F' && strncmp(buf2,"From ",5)==0)
-	      putc('>',fp);
-	    i = fwrite(buf2, 1, tlen, fp);
-	    if (i != tlen) return -1;
-	    return tlen;
-	  }
-	  /* ------------------------------------------------------------ */
-	  /* Well, no other processings known.. */
-	  if (!mmdf_mode && buf[0] == 'F' && strncmp(buf,"From ",5)==0)
-	    putc('>',fp);
-	  i = fwrite(buf, 1, len, fp);
-	  if (i != len) return -1;
-	  return len;
-	}
-
-	/* ------------------------------------------------------------ */
-	/* Old-fashioned processing */
+	/* -------------------------------------------- */
+	/* Non-MIME processing				*/
 
 	expect = save;
 
@@ -1731,6 +1705,108 @@ writebuf(fp, buf, len, ismime)
 	return -1;	/* errno is correct if tlen == -1 ! */
 }
 
+int
+writemimebuf(fp, buf, len)
+	FILE *fp;
+	char *buf;
+	int len;
+{
+	register char *cp, *s;
+	register int n;
+	int wlen, tlen, i = 0;
+	register char expect;
+	static char save = '\0';
+	static char frombuf[8];
+	static char *fromp = NULL;
+
+	static char *buf2 = NULL;
+	static int buf2len = 0;
+
+	if (buf == NULL) {	/* magic initialization */
+	  save = mmdf_mode ? 0 : 'F';
+	  frombuf[0] = 0;
+	  fromp = frombuf;
+	  return 0;
+	}
+
+#ifdef	USE_ALLOCA
+	buf2 = alloca(len+1);
+#else
+	if (buf2 == NULL) {
+	  buf2 = emalloc(len+1);
+	  buf2len = len;
+	}
+	if (buf2len < len) {
+	  free(buf2);
+	  buf2 = emalloc(len+1);
+	  buf2len = len;
+	}
+#endif
+
+	/* ------------------------------------------------------------ */
+	/* MIME format processing, the input is LINE BY LINE */
+	if (mime_boundary != NULL && n > mime_boundary_len+2 &&
+	    buf[0] == '-' && buf[1] == '-' &&
+	    strncmp(buf+2,mime_boundary,mime_boundary_len) == 0) {
+	  /* ------------------------------------------------------------ */
+	  /* XX: process MIME boundary! */
+	  i = fwrite(buf, 1, len, fp);
+	  if (i != len) return -1;
+	  return len;
+	} else if (convert_qp) {
+	  /* ------------------------------------------------------------ */
+	  /* Process Q-P chars of this line */
+	  int qp_hex = 0;
+	  int qp_chrs = 0;
+	  int tlen = 0;
+
+	  for (cp = buf, s = buf2, n = len; n > 0; --n, ++cp) {
+	    register int c = *cp;
+	    /* At first, convert the possible QP character.. */
+	    if (!qp_chrs && c == '=') {
+	      qp_chrs = 2;
+	      qp_hex  = 0;
+	      continue;		/* '=' starts something */
+	    }
+	    if (qp_chrs && c == '\n') {
+	      qp_chrs = 0;	/* It was "=[ \t]*\n", which was thrown away.*/
+	      continue;
+	    }
+	    if (qp_chrs == 2 && (c == ' ' || c == '\t'))
+	      continue;		/* It is "=[ \t]*\n", throw it away */
+	    /* Ok, this may be of  =HexHex ? */
+	    if (qp_chrs && ((c >= '0' && c <= '9') ||
+			    (c >= 'A' && c <= 'F') ||
+			    (c >= 'a' && c <= 'f'))) {
+	      qp_hex <<= 4;
+	      if (c >= '0' && c <= '9') qp_hex += (c - '0');
+	      if (c >= 'A' && c <= 'F') qp_hex += (c - 'A' + 10);
+	      if (c >= 'a' && c <= 'f') qp_hex += (c - 'a' + 10);
+	      --qp_chrs;
+	      if (!qp_chrs) c = qp_hex;
+	      else continue;	/* Go collect another digit */
+	    } else if (qp_chrs)
+	      qp_chrs = 0;	/* Failed a HEX check.. */
+	    /* Now we have a converted character at `c', save it! */
+	    *s++ = c;
+	    ++tlen;
+	  }
+	  /* Uhh... Stored the conversion result into buf2[], length  tlen */
+	  if (!mmdf_mode && buf2[0] == 'F' && strncmp(buf2,"From ",5)==0)
+	    putc('>',fp);
+	  i = fwrite(buf2, 1, tlen, fp);
+	  if (i != tlen) return -1;
+	  return len;		/* Return the incoming length, NOT the true length! */
+	}
+	/* ------------------------------------------------------------ */
+	/* Well, no other processings known.. */
+	if (!mmdf_mode && buf[0] == 'F' && strncmp(buf,"From ",5)==0)
+	  putc('>',fp);
+	i = fwrite(buf, 1, len, fp);
+	if (i != len) return -1;
+	return len;
+}
+
 
 /* [mea] See if we have a MIME  TEXT/PLAIN message encoded with
          QUOTED-PRINTABLE... */
