@@ -1,6 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Copyright 1994 by Matti Aarnio -- MIME processings
  */
 
 #include "hostenv.h"
@@ -54,16 +55,15 @@ extern char *optarg;
 extern int optind;
 extern int getmyuucpname(), emptyline();
 extern struct maildesc *readsmcf();
-extern void prversion(), process();
-extern void process(), deliver();
-extern int appendlet();
-extern int writebuf();
+extern void prversion();
+extern void process __((struct ctldesc *dp, struct maildesc *mp, FILE *verboselog));
+extern void deliver __((struct ctldesc *dp, struct maildesc *mp, struct rcpt *startrp, struct rcpt *endrp, FILE *verboselog));
+extern int appendlet __((struct ctldesc *dp, struct maildesc *mp, FILE *fp, FILE *verboselog, int convertmode));
+extern int writebuf __((struct maildesc *mp, FILE *fp, char *buf, int len));
 extern time_t time();
 extern char *strchr(), *strrchr();
 
-extern int  cte_check __((struct rcpt *rp));
-extern void downgrade_headers __((struct rcpt *rp, int downgrade, FILE *verboselog));
-extern void check_7bit_downgrade __((int mailfd, int *flag_ptr));
+extern int check_7bit_cleanness __((struct ctldesc *dp));
 
 struct maildesc {
 	char	*name;
@@ -225,16 +225,13 @@ process(dp, mp, verboselog)
 	    if (rp->next == NULL
 		|| rp->addr->link != rp->next->addr->link
 		|| rp->newmsgheader != rp->next->newmsgheader) {
-	      deliver(mp, rphead, rp->next,
-		      dp->msgfd, dp->msgbodyoffset,
-		      verboselog);
+	      deliver(dp, mp, rphead, rp->next, verboselog);
 	      rphead = rp->next;
 	    }
 	  }
 	} else {
 	  for (rp = dp->recipients; rp != NULL; rp = rp->next)
-	    deliver(mp, rp, rp->next, dp->msgfd, dp->msgbodyoffset,
-		    verboselog);
+	    deliver(dp, mp, rp, rp->next, verboselog);
 	}
 }
 
@@ -244,11 +241,10 @@ process(dp, mp, verboselog)
  */
 
 void
-deliver(mp, startrp, endrp, messagefd, msgoffset, verboselog)
+deliver(dp, mp, startrp, endrp, verboselog)
+	struct ctldesc *dp;
 	struct maildesc *mp;
 	struct rcpt *startrp, *endrp;
-	int messagefd;
-	long msgoffset;
 	FILE *verboselog;
 {
 	struct rcpt *rp;
@@ -256,17 +252,19 @@ deliver(mp, startrp, endrp, messagefd, msgoffset, verboselog)
 	int i, j, pid, in[2], out[2];
 	unsigned int avsize;
 	FILE *tafp, *errfp;
-	char *cp, *dp, *ds, *s, buf[BUFSIZ], buf2[BUFSIZ];
+	char *cp, *ds, *s, buf[BUFSIZ], buf2[BUFSIZ];
 	char **av;
 #ifdef	USE_UNIONWAIT
 	union wait status;
 #else	/* !USE_UNIONWAIT */
 	int	status;
 #endif	/* USE_UNIONWAIT */
-	int content_kind, downgrade;
+	int content_kind, conversion_prohibited, convertmode, ascii_clean = 0;
 
-	if (lseek(messagefd, msgoffset, L_SET) < 0L)
+#ifndef USE_MALLOC
+	if (lseek(dp->msgfd, dp->msgbodyoffset, L_SET) < 0L)
 		warning("Cannot seek to message body! (%m)", (char *)NULL);
+#endif
 	i = 0;
 	avsize = 5;
 	av = (char **)emalloc(sizeof av[0] * avsize);
@@ -310,9 +308,9 @@ deliver(mp, startrp, endrp, messagefd, msgoffset, verboselog)
 		  break;
 		case 'U':
 		  strcpy(buf2, rp->addr->user);
-		  for (dp = buf2; *dp != 0; ++dp)
-		    if (isupper(*dp))
-		      *dp = tolower(*dp);
+		  for (ds = buf2; *ds != 0; ++ds)
+		    if (isupper(*ds))
+		      *ds = tolower(*ds);
 		  ds = buf2;
 		  rp = rp->next;
 		  break;
@@ -414,18 +412,53 @@ deliver(mp, startrp, endrp, messagefd, msgoffset, verboselog)
 	  putc('\n', tafp);
 	}
 
+	conversion_prohibited = check_conv_prohibit(startrp);
+
 	/* Content-Transfer-Encoding: 8BIT ? */
 	content_kind = cte_check(startrp);
-	downgrade = (content_kind  == 8) && !can_8bit;
-	/* If it does not contain MIME headers, and it contains 8-bit
-	   chars, it is in trouble!  Mark the checkup with  -1       */
-	if (content_kind == 0 && !can_8bit)
-	  downgrade = -1;
-	if (downgrade) {
-	  check_7bit_downgrade(messagefd, &downgrade);
-	  downgrade_headers(startrp, downgrade, verboselog);
+
+	/* If the header says '8BIT' and ISO-8859-* something,
+	   but body is plain 7-bit, turn it to '7BIT', and US-ASCII */
+	ascii_clean = check_7bit_cleanness(dp);
+	if (!conversion_prohibited && ascii_clean && content_kind == 8) {
+	  if (downgrade_charset(startrp, verboselog))
+	    content_kind = 7;
 	}
 
+	convertmode = _CONVERT_NONE;
+	if (!conversion_prohibited)
+	  switch (content_kind) {
+	  case 0:		/* No MIME headers defined */
+	    if (!can_8bit && !ascii_clean) {
+	      convertmode = _CONVERT_UNKNOWN;
+	      downgrade_headers(startrp, convertmode, verboselog);
+	    }
+	    break;
+	  case 2:		/* MIME, but no C-T-E: ? */
+	  case 1:		/* MIME BASE64 ? some MIME anyway.. */
+	  case 7:		/* 7BIT */
+	    convertmode = _CONVERT_NONE;
+	    break;
+	  case 8:		/* 8BIT */
+	    if (!can_8bit && !ascii_clean) {
+	      convertmode = _CONVERT_QP;
+	      downgrade_headers(startrp, convertmode, verboselog);
+	    }
+	    break;
+	  case 9:		/* QUOTED-PRINTABLE */
+	    if (can_8bit) {
+	      /* Force(d) to decode Q-P while transfer.. */
+	      convertmode = _CONVERT_8BIT;
+	      /*  UPGRADE TO 8BIT !  */
+	      qp_to_8bit(startrp);
+	      content_kind = 10;
+	      ascii_clean = 0;
+	    }
+	    break;
+	  default:		/* ?? should not happen.. */
+	    break;
+	  }
+
  	/* Add the "Return-Path:" is it is desired, but does not yet
 	   exist.. */
 	if (mp->flags & MO_RETURNPATH)
@@ -433,12 +466,12 @@ deliver(mp, startrp, endrp, messagefd, msgoffset, verboselog)
 	    append_header(startrp,"Return-Path: <%s>\n",
 			  startrp->addr->link->user);
 	if (mp->flags & MO_CRLF)
-	  writeheaders(startrp,tafp, "\r\n",downgrade & 1, maxwidth);
+	  writeheaders(startrp, tafp, "\r\n", convertmode, maxwidth);
 	else
-	  writeheaders(startrp,tafp, "\n",  downgrade & 1, maxwidth);
+	  writeheaders(startrp, tafp, "\n",   convertmode, maxwidth);
 
 	/* append message body itself */
-	if ((i = appendlet(mp, tafp, messagefd, verboselog, downgrade)) != EX_OK) {
+	if ((i = appendlet(dp, mp, tafp, verboselog, convertmode)) != EX_OK) {
 	  for (rp = startrp; rp != endrp; rp = rp->next)
 	    diagnostic(rp, i, "write error");
 	  /* just to make sure nothing will get delivered */
@@ -513,48 +546,127 @@ deliver(mp, startrp, endrp, messagefd, msgoffset, verboselog)
  * appendlet - append letter to file pointed at by fd
  */
 
+#ifndef	USE_MMAP
 static char let_buffer[BUFSIZ];
+#endif
 
 int
-appendlet(mp, fp, mfd, verboselog, downgrade)
+appendlet(dp, mp, fp, verboselog, convertmode)
+	struct ctldesc *dp;
 	struct maildesc *mp;
 	FILE *fp;
-	int mfd;
 	FILE *verboselog;
-	int downgrade;
+	int convertmode;
 {
+	/* `convertmode' controls the behaviour of the message conversion:
+	     _CONVERT_NONE (0): send as is
+	     _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
+	     _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
+	     _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
+	 */
+
 	register int i;
 	register int bufferfull;
 	int lastwasnl;
 
-	writebuf(mp, fp, (char *)NULL, 0, 0);  /* magic initialization */
+	writebuf(mp, fp, (char *)NULL, 0);  /* magic initialization */
 
+#ifndef USE_MMAP
 	/* can we use cache of message body data */
-	if (readalready != 0) {
-	  if (writebuf(mp, fp, let_buffer, readalready, downgrade) != readalready)
+	if (convertmode == _CONVERT_NONE &&
+	    readalready != 0) {
+	  if (writebuf(mp, fp, let_buffer, readalready) != readalready)
 	    return EX_IOERR;
 	  if (let_buffer[readalready-1] != '\n')
-	    writebuf(mp, fp, "\n", 1, 0);
+	    if (writebuf(mp, fp, "\n", 1) != 1)
+	      return EX_IOERR;
 	  return EX_OK;
 	}
+#endif
 
 	/* we are assumed to be positioned properly at start of message body */
 	bufferfull = lastwasnl = 0;
-	while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
-	  if (i < 0)
+	if (convertmode == _CONVERT_NONE) {
+#ifndef	USE_MMAP
+	  while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
+#else
+	    char *let_buffer = dp->let_buffer + dp->msgbodyoffset;
+	    i = dp->let_end - dp->let_buffer - dp->msgbodyoffset;
+#endif
+	    if (i < 0)
+	      return EX_IOERR;
+	    lastwasnl = (let_buffer[i-1] == '\n');
+	    if (writebuf(mp, fp, let_buffer, i) != i)
+	      return EX_IOERR;
+	    readalready = i;
+	    bufferfull++;
+#ifndef	USE_MMAP
+	  }
+#endif
+	} else {
+	  /* convertmode something else, than _CONVERT_NONE */
+	  /* Various esoteric conversion modes..
+	     We are better to feed writemimeline() with LINES
+	     instead of blocks of data.. */
+#ifndef	USE_MMAP
+	  char iobuf[BUFSIZ];
+	  FILE *mfp = fdopen(mfd,"r");
+
+	  writemimeline(mp, fp, (char *)NULL, 0, 0);
+	  setvbuf(mfp, iobuf, _IOFBF, sizeof(iobuf));
+
+#define MFPCLOSE i = dup(mfd); fclose(mfp); dup2(i,mfd); close(i);
+
+	  readalready = 0;
+#else
+#define MFPCLOSE
+	  char *s = dp->let_buffer + dp->msgbodyoffset;
+#endif
+
+	  /* we are assuming to be positioned properly
+	     at the start of the message body */
+	  bufferfull = 0;
+	  lastwasnl = 0;
+	  i = 0;
+	  for (;;) {
+#ifndef USE_MMAP
+	    if ((i = cfgets(let_buffer, sizeof(let_buffer), mfp)) == EOF)
+	      break;
+	    lastwasnl = (let_buffer[i-1] == '\n');
+#else
+	    char *let_buffer = s, *s2 = s;
+	    if (s >= dp->let_end) break;	/* "EOF" */
+	    i = 0;
+	    while (s2 < dp->let_end && *s2 != '\n')
+	      ++s2, ++i;
+	    if ((lastwasnl = (*s2 == '\n')))
+	      ++s2, ++i;
+	    s = s2;
+#endif
+	    /* It MAY be malformed -- if it has a BUFSIZ length
+	       line in it, IT CAN'T BE MIME  :-/		*/
+	    /* Ok, write the line */
+	    if (writemimeline(mp, fp, let_buffer, i, convertmode) != i) {
+	      return EX_IOERR;
+	    }
+	  }
+#ifndef USE_MMAP
+	  if (i == EOF && !feof(mfp)) {
+	    MFPCLOSE
 	    return EX_IOERR;
-	  lastwasnl = (let_buffer[i-1] == '\n');
-	  if (writebuf(mp, fp, let_buffer, i, downgrade) != i)
-	    return EX_IOERR;
-	  readalready = i;
-	  bufferfull++;
+	  }
+	  MFPCLOSE
+#endif
 	}
+
 	/* we must make sure the last thing we transmit is a CRLF sequence */
 	if (!lastwasnl)
-	  writebuf(mp, fp, "\n", 1, downgrade);
+	  writebuf(mp, fp, "\n", 1);
 
+#ifndef USE_MMAP
 	if (bufferfull > 1)	/* not all in memory, need to reread */
 	  readalready = 0;
+#endif
 	return EX_OK;
 }
 
@@ -564,56 +676,36 @@ appendlet(mp, fp, mfd, verboselog, downgrade)
  */
 
 int
-writebuf(mp, fp, buf, len, downgrade)
+writebuf(mp, fp, buf, len)
 	struct maildesc *mp;
 	FILE *fp;
 	char *buf;
 	int len;
-	int downgrade;
 {
 	register char *cp;
 	register int n;
 	int tlen;
 	register char expect;
-	register int  column;
 	static char save = '\0';
-	static int  col  = 0;
 	static char frombuf[8];
 	static char *fromp;
 
 	if (buf == NULL) {	/* magic initialization */
 	  save = '.';
-	  col  = 0;
 	  frombuf[0] = 0;
 	  fromp = frombuf;
 	  return 0;
 	}
 	expect = save;
-	column = col;
-	downgrade = (downgrade & 1) != 0;
-	if ((mp->flags & MO_STRIPHIBIT) && !downgrade) {
+	if ((mp->flags & MO_STRIPHIBIT)) {
 	  for (cp = buf, n = len; n > 0; --n, ++cp)
 	    if (*cp & 0200)
 	      *cp &= 0177;
 	}
 	for (cp = buf, n = len, tlen = 0; n > 0; --n, ++cp) {
 	  int c = *(unsigned char *)cp;
-	  ++column;
 	  ++tlen;
-	  if (downgrade && column > 70 && c != '\n') {
-	    putc('=',fp);
-	    if (mp->flags & MO_CRLF)
-	      putc('\r', fp);
-	    putc('\n',fp);
-	    column = 1;
-	    expect = '.';
-	  }
-	  if (downgrade && c != '\n' && (c < 32 || c > 126)) {
-	    /* Downgrade it by translating it to Quoted-Printable.. */
-	    fprintf(fp,"=%02X",c);
-	    column += 2;
-	    buf = cp;
-	  } else if (c == '\n') {
+	  if (c == '\n') {
 	    frombuf[0] = 0;
 	    fromp = frombuf;
 	    if (expect == '\n' && (mp->flags & MO_HIDDENDOT))
@@ -623,7 +715,6 @@ writebuf(mp, fp, buf, len, downgrade)
 	      if (putc('\r', fp) == EOF) { tlen = -1; break; }
 	    if (putc(c,fp) == EOF) { tlen = -1; break; }
 	    expect = '.';
-	    column = 0;
 	  } else if (expect != '\0') {
 	    if (expect == '.') {
 	      if ((mp->flags & MO_ESCAPEFROM) && c == 'F')
@@ -654,7 +745,6 @@ writebuf(mp, fp, buf, len, downgrade)
 		    { tlen = -1; break; }
 		  /* anticipate future instances */
 		  expect = '\0';
-		  ++column;
 		  break;
 	      }
 	    } else {
@@ -676,11 +766,118 @@ writebuf(mp, fp, buf, len, downgrade)
 	  }
 	}
 	save = expect;
-	col  = column;
 
 	return tlen;
 }
 
+int
+writemimeline(mp, fp, buf, len, convertmode)
+	struct maildesc *mp;
+	FILE *fp;
+	char *buf;
+	int len, convertmode;
+{
+	/* `convertmode' controls the behaviour of the message conversion:
+	     _CONVERT_NONE (0): send as is
+	     _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
+	     _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
+	     _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
+	 */
+	register char *cp;
+	register int n;
+	register char expect;
+	static int  column;
+	register int qp_conv;
+	register int qp_chrs = 0;
+
+	if (buf == NULL) {
+	  column = -1;
+	  return 0;
+	}
+
+	qp_conv = (convertmode == _CONVERT_QP ||
+		   convertmode == _CONVERT_UNKNOWN);
+
+	for (cp = buf, n = len; n > 0; --n, ++cp) {
+	  int c = *(unsigned char *)cp;
+	  ++column;
+
+	  if (qp_conv) {
+	    /* ENCODE to QUOTED-PRINTABLE ... */ 
+	    if (column > 70 && c != '\n') {
+	      putc('=',fp);
+	      if (mp->flags & MO_CRLF)
+		putc('\r', fp);
+	      putc('\n',fp);
+	      column = 0;
+	    }
+
+	    if (column == 0 && (mp->flags & MO_HIDDENDOT) && c == '.') {
+	      /* Duplicate the line initial dot.. */
+	      if (putc(c,fp)==EOF) return EOF;
+	    } else if (column == 0  &&  (mp->flags & MO_ESCAPEFROM)  &&
+		       c == 'F'  &&  n >= 4  &&  strncmp(cp,"From",4)==0) {
+	      /* We Q-P encode the leading 'F'.. */
+	      if (fputs("=46",fp) != 3) return EOF;
+	      column += 2;
+	    } else if (c != '\n' && (c < 32 || c > 126 || c == '=')) {
+	      /* Downgrade it by translating it to Quoted-Printable.. */
+	      fprintf(fp,"=%02X",c);
+	      column += 2;
+	      buf = cp;
+	    } else if (c == '\n') { /* This is most likely the LAST char */
+	      if (mp->flags & MO_CRLF)
+		if (putc('\r', fp) == EOF) return EOF;
+	      if (putc(c,fp) == EOF) return EOF;
+	      column = -1;
+	    } else if (expect != '\0') {
+	      if (putc(c, fp) == EOF) return EOF;
+	    }
+	  } else if (convertmode == _CONVERT_8BIT) {
+	    /* DECODE from QUOTED-PRINTABLE text.. */
+	    static int qp_val = 0;
+	    if (!qp_chrs && c == '=') { /* Q-P -prefix */
+	      qp_chrs = 2;
+	      qp_val = 0;
+	      continue;
+	    } else if (qp_chrs) {
+	      --column;
+	      if (c == ' ' || c == '\t' || c == '\n' || c == '\r')
+		break; /* Done with it, it was soft end-of-line */
+	      if ((c >= '0' && c <= '9') || (c >= 'A' && c <= 'F') ||
+		  (c >= 'a' && c <= 'f')) {
+		/* A HEX digit ? QP char coming up ? */
+		if (c >= 'a') c -= ('a' - 'A');
+		if (c >= 'A') c -= ('A' - '9' -1);
+		qp_val <<= 4;
+		qp_val |= (c & 0x0F);
+		--qp_chrs;
+	      }
+	      if (!qp_chrs)
+		c = qp_val;
+	    } /* Ok, decoded possible Q-P chars.  Now normal processing.. */
+	    if (column == 0 && c == '.' && (mp->flags & MO_HIDDENDOT)) {
+	      if (putc(c,fp)==EOF) return EOF;
+	    } else if (column == 0 && (mp->flags & MO_ESCAPEFROM) &&
+		       c == 'F' && strncmp(cp+1,"rom",3)==0) {
+	      if (putc('>',fp)==EOF) return EOF;
+	      ++column;
+	    } else if (c == '\n') {
+	      if (mp->flags & MO_CRLF)
+		if (putc('\r',fp)==EOF) return EOF;
+	      column = -1;
+	    }
+	    /* And output the char.. */
+	    if (putc(c,fp)==EOF) return EOF;
+	  } else
+	    abort(); /* WOO! We should not be called for '_CONVERT_NONE'! */
+	}
+
+	if (feof(fp) || ferror(fp)) return EOF;
+	return len;
+}
+
+
 struct maildesc *
 readsmcf(file, mailer)
 	char *file, *mailer;
@@ -817,12 +1014,19 @@ readsmcf(file, mailer)
 	return &m;
 }
 
-/* When data is clean 7-BIT, do:  *flag_ptr = (*flag_ptr) << 1  */
-void
-check_7bit_downgrade(mfd, flag_ptr)
-int *flag_ptr;
-int mfd;
+/* When data is clean 7-BIT ...  */
+int
+check_7bit_cleanness(dp)
+struct ctldesc *dp;
 {
+#ifdef	USE_MMAP
+	/* With MMAP()ed spool file it is sweet and simple.. */
+	register char *s = dp->let_buffer + dp->msgbodyoffset;
+	while (s < dp->let_end)
+	  if (128 & *s) return 0;
+	  else ++s;
+	return 1;
+#else
 	register int i;
 	register int bufferfull;
 	int lastwasnl;
@@ -832,19 +1036,17 @@ int mfd;
 	if (readalready != 0) {
 	  for (i=0; i<readalready; ++i)
 	    if (128 & (let_buffer[i]))
-	      return;		/* Not clean ! */
+	      return 1;		/* Not clean ! */
 	}
 
 	/* we are assumed to be positioned properly at start of message body */
 	bufferfull = 0;
 
-	mfd_pos = tell(mfd);
-	
 	while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
 	  if (i < 0) {
 	    /* ERROR ?!?!? */
 	    if (bufferfull > 1) readalready = 0;
-	    return;
+	    return 1;
 	  }
 	  lastwasnl = (let_buffer[i-1] == '\n');
 	  readalready = i;
@@ -853,14 +1055,14 @@ int mfd;
 	    if (128 & (let_buffer[i])) {
 	      lseek(mfd,mfd_pos,0);
 	      if (bufferfull > 1) readalready = 0;
-	      return;		/* Not clean ! */
+	      return 1;		/* Not clean ! */
 	    }
 	}
 	/* Got to EOF, and still it is clean 7-BIT! */
-	lseek(mfd,mfd_pos,0);
+	lseek(mfd,dp->msgbodyoffset,0);
 	if (bufferfull > 1)	/* not all in memory, need to reread */
 	  readalready = 0;
 
-	/* Set indication! */
-	*flag_ptr = (*flag_ptr) << 1;
+	return 0;
+#endif
 }
