@@ -1,6 +1,7 @@
 /*
  *	Copyright 1988 by Rayan S. Zachariassen, all rights reserved.
  *	This will be free software, but only when it is finished.
+ *	Copyright 1991-1994 by Matti Aarnio -- modifications, including MIME
  */
 
 #define	RFC974		/* If BIND, check that TCP SMTP service is enabled */
@@ -29,6 +30,13 @@
 #include "ta.h"
 #include "malloc.h"
 
+#ifndef	USE_SETREUID
+#ifdef	USE_SETEUID
+#define	setreuid(x,y)	seteuid(y)
+#else
+#define	setreuid(x,y)	setuid(y)
+#endif	/* !USE_SETEUID */
+#endif	/* !USE_SETREUID */
 
 #if	defined(TRY_AGAIN) && defined(USE_RESOLV)
 #define	BIND		/* Want BIND (named) nameserver support enabled */
@@ -54,26 +62,39 @@ typedef u_char msgdata;
 typedef char msgdata;
 #endif	/* defined(BIND_VER) && (BIND_VER >= 473) */
 
+#ifdef	__alpha	/* This is ugly kludge due to the nature of the IP
+		   addresses.. They are 32 bit entities, not 64 bit! */
+# define  u_long u_int
+#endif
 
 /* Define all those things which exist on newer BINDs, and which may
    get returned to us, when we make a query with  T_ANY ... */
 
 #ifndef	T_TXT
-# define T_TXT 16
+# define T_TXT 16	/* Text strings */
 #endif
 #ifndef T_RP
-# define T_RP
+# define T_RP 17	/* Responsible person */
 #endif
 #ifndef T_AFSDB
-# define T_AFSDB 18
+# define T_AFSDB 18	/* AFS cell database */
+#endif
+#ifndef T_X25
+# define T_X25 19	/* X.25 calling address */
+#endif
+#ifndef T_ISDN
+# define T_ISDN 20	/* ISDN calling address */
+#endif
+#ifndef T_RT
+# define T_RT 21	/* router */
 #endif
 #ifndef T_NSAP
-# define T_NSAP 22
+# define T_NSAP 22	/* NSAP address */
 #endif
 #ifndef T_NSAP_PTR
-# define T_NSAP_PTR 23
+# define T_NSAP_PTR 23	/* reverse NSAP lookup (depreciated) */
 #endif
-#ifndef	T_UINFO		/* Defined on several newer BINDs.. */
+#ifndef	T_UINFO
 # define T_UINFO 100
 #endif
 #ifndef T_UID
@@ -86,7 +107,7 @@ typedef char msgdata;
 # define T_UNSPEC 103
 #endif
 #ifndef T_SA
-# define T_SA 200
+# define T_SA 200		/* Shuffle addresses */
 #endif
 
 
@@ -94,6 +115,12 @@ typedef char msgdata;
 #ifndef	L_SET
 #define	L_SET	0
 #endif	/* L_SET */
+#ifndef L_CUR
+#define L_CUR   1
+#endif
+#ifndef L_XTND
+#define L_XTND  2
+#endif
 
 #ifndef	IPPORT_SMTP
 #define	IPPORT_SMTP	25
@@ -136,14 +163,24 @@ int conndebug = 0;
 int timeout = 0;		/* how long do we wait for response? (sec.) */
 int gotalarm =0;		/* indicate that alarm happened! */
 #endif
+#ifndef	USE_MMAP
 int readalready = 0;		/* does buffer contain valid message data? */
+#endif
 int wantreserved = 0;		/* open connection on secure (reserved) port */
 int statusreport = 0;		/* put status reports on the command line */
 int noMX = 0;			/* do not use MX records to deliver mail */
 int firstmx = 0;		/* [Matti Aarnio] error in smtpwrite("HELO".. */
 int mxcount = 0;
+int force_8bit = 0;		/* Claim to the remote to be 8-bit system, even
+				   when it doesn't report itself as such..*/
+int force_7bit = 0;		/* and reverse the previous.. */
 int checkwks = 0;
 FILE *logfp;
+extern int nobody;
+int daemon_uid = -1;
+int first_uid = 0;		/* Make the opening connect with the UID of the
+				   sender (atoi(rp->addr->misc)), unless it is
+				   "nobody", in which case use "daemon"      */
 
 int D_alloc = 0;		/* Memory usage debug */
 
@@ -152,8 +189,12 @@ int D_alloc = 0;		/* Memory usage debug */
 #define	putc	fputc
 #endif	/* lint */
 
-/* Extended SMTP flags -- can downgrade from 8-bit to 7-bit while at transport
-   IF  MIME-Version: is present, AND Content-Transfer-Encoding: 8BIT */
+/* Extended SMTP flags -- can downgrade from 8-bit to 7-bit while in transport
+   IF  MIME-Version: is present, AND Content-Transfer-Encoding: 8BIT
+   For selected "force_8bit" remotes can also DECODE Q-P MIME MSGS! */
+/* If there is header:  Content-Conversion: prohibited
+   DO NOT do conversions no matter what
+   (even when it violates the protocol..) */
 
 /* Following options can be declared in ESMTP  EHLO responce  */
 #define ESMTP_SIZEOPT  0x0001
@@ -176,9 +217,10 @@ extern char *strrchr();
 extern char *dottedquad();
 extern char *optarg;
 extern int optind;
-extern int deliver();
-extern int writebuf __((FILE *fp, char *buf, int len, int downgrade));
-extern int appendlet __((FILE *fp, FILE *verboselog, int mfp, int hsize, int msize, int downgrade));
+extern int deliver __((struct ctldesc *dp, FILE *smtpfp[2], FILE *verboselog, struct rcpt *startrp, struct rcpt *endrp));
+extern int writebuf __((FILE *fp, char *buf, int len));
+extern int writemimeline __((FILE *fp, char *buf, int len, int convertmode));
+extern int appendlet __((struct ctldesc *dp, FILE *fp, FILE *verboselog, int hsize, int msize, int convertmode));
 extern int smtpopen __((char *host, FILE *fparr[], FILE *verboselog));
 extern int smtpconn __((char *host, FILE *fparr[], FILE *verboselog));
 extern int smtp_ehlo __((FILE *smtpfp[], FILE *verboselog, char *strbuf));
@@ -191,10 +233,11 @@ extern void notarystatsave __((char *smtpstatline));
 extern char *notaryacct __((int rc, char *ok_string));
 
 extern int  cte_check __((struct rcpt *rp));
-extern void downgrade_headers __((struct rcpt *rp, int downgrade, FILE *verboselog));
-extern void check_7bit_downgrade __((int mailfd, int *flag_ptr));
+extern void downgrade_headers __((struct rcpt *rp, int convertmode, FILE *verboselog));
+extern int check_7bit_cleanness __((struct ctldesc *dp));
 
 extern void prversion();
+extern void getnobody();
 extern int makeconn();
 extern int vcsetup();
 extern void hp_init();
@@ -217,7 +260,7 @@ extern SIGNAL_TYPE sig_alarm();
 extern int getmyhostname();
 extern void stashmyaddresses();
 extern int cistrcmp(), cistrncmp();
-
+extern void getdaemon();
 
 FILE *verboselog = NULL;
 
@@ -268,7 +311,7 @@ main(argc, argv)
 	myhostname[0] = '\0';
 	remotemsg[0] = '\0';
 	remotehost[0] = '\0';
-	while ((c = getopt(argc, argv, "c:deh:l:rsxDE:T:VW")) != EOF) {
+	while ((c = getopt(argc, argv, "c:deh:l:rsxDE:T:VW78")) != EOF) {
 	  switch (c) {
 	  case 'c':		/* specify channel scanned for */
 	    channel = emalloc(strlen(optarg)+1);
@@ -309,6 +352,7 @@ main(argc, argv)
 	  case 'E':		/* don't do EHLO, unless target system
 				   has "ESMTP" on its banner */
 	    esmtp_on_banner = 0;
+	    break;
 	  case 'T':		/* specify Timeout in seconds */
 #ifdef	TIMEOUT
 	    if ((timeout = atoi(optarg)) < 5) {
@@ -330,6 +374,13 @@ main(argc, argv)
 	  case 'W':		/* Enable RFC974 WKS checks */
 	    checkwks = 1;
 	    break;
+	  case '8':
+	    force_8bit = 1;
+	    force_7bit = 0;
+	    break;
+	  case '7':
+	    force_7bit = 1;
+	    force_8bit = 0;
 	  default:
 	    ++errflg;
 	    break;
@@ -366,6 +417,10 @@ main(argc, argv)
 	} else
 		logfp = NULL;
 
+	getnobody();
+	getdaemon();
+
+
 	/* defer opening a connection until we know there is work */
 	smtpstatus = EX_DATAERR;
 	while (!getout && fgets(file, sizeof file, stdin)) {
@@ -388,8 +443,12 @@ main(argc, argv)
 	    if (verboselog)
 	      setbuf(verboselog,NULL);
 	  }
-	  if (smtpstatus == EX_DATAERR)
+	  if (smtpstatus == EX_DATAERR) {
+	    if ((first_uid = atoi(dp->senders->misc)) < 0 ||
+		first_uid == nobody)
+	      first_uid = daemon_uid;
 	    smtpstatus = smtpopen(host, smtpfp, verboselog);
+	  }
 	  smtpstatus = process(dp, smtpstatus, host, smtpfp, verboselog);
 	  ctlclose(dp);
 	}
@@ -417,7 +476,9 @@ process(dp, smtpstatus, host, smtpfp, verboselog)
 	struct rcpt *rp, *rphead;
 	int loggedid;
 
+#ifndef	USE_MMAP
 	readalready = 0; /* ignore any previous message data cache */
+#endif
 	loggedid = 0;
 
 	for (rp = rphead = dp->recipients; rp != NULL; rp = rp->next) {
@@ -433,11 +494,13 @@ process(dp, smtpstatus, host, smtpfp, verboselog)
 			dp->logident);
 		fflush(logfp);
 	      }
-	      smtpstatus = deliver(smtpfp, verboselog, rphead, rp->next,
-				   dp->msgfd, dp->msgbodyoffset);
+	      smtpstatus = deliver(dp, smtpfp, verboselog, rphead, rp->next);
 	      if (smtpstatus != EX_OK) {
 		fclose(smtpfp[0]);
 		fclose(smtpfp[1]);
+		if ((first_uid = atoi(dp->senders->misc)) < 0 ||
+		    first_uid == nobody)
+		  first_uid = daemon_uid;
 		smtpstatus = smtpopen(host, smtpfp, verboselog);
 	      }
 	      rphead = rp->next;
@@ -458,37 +521,91 @@ process(dp, smtpstatus, host, smtpfp, verboselog)
  */
 
 int
-deliver(smtpfp, verboselog, startrp, endrp, messagefd, msgoffset)
+deliver(dp, smtpfp, verboselog, startrp, endrp)
+	struct ctldesc *dp;
 	FILE *smtpfp[2], *verboselog;
 	struct rcpt *startrp, *endrp;
-	int messagefd;
-	long msgoffset;
 {
-	struct rcpt *rp;
+	struct rcpt *rp = NULL;
 	int r, nrcpt, hsize, size, tout;
-	int content_kind, flag_downgrade;
+	int content_kind, convertmode;
+	int ascii_clean = 0;
 	struct stat stbuf;
 	char SMTPbuf[2000];
 	char *s;
+	int conv_prohibit = check_conv_prohibit(startrp);
 
-	if (lseek(messagefd, msgoffset, L_SET) < 0L)
-	  warning("Cannot seek to message body! (%m)", (char *)NULL);
-	sprintf(SMTPbuf,"MAIL From:<%s>",
-		startrp->addr->link->user);
+	if (conv_prohibit) {
+	  convertmode = _CONVERT_NONE;
+	} else {
+
+	  /* Content-Transfer-Encoding: 8BIT ? */
+	  content_kind = cte_check(startrp);
+
+	  /* If the header says '8BIT' and ISO-8859-* something,
+	     but body is plain 7-bit, turn it to '7BIT', and US-ASCII */
+	  ascii_clean = check_7bit_cleanness(dp);
+	  if (ascii_clean && content_kind == 8) {
+	    if (downgrade_charset(startrp, verboselog))
+	      content_kind = 7;
+	  }
+
+	  if (force_7bit)	/* Mark off the 8BIT MIME capability.. */
+	    ehlo_capabilities &= ~ESMTP_8BITMIME;
+
+	  convertmode = _CONVERT_NONE;
+	  switch (content_kind) {
+	  case 0:		/* Not MIME */
+	    if ((ehlo_capabilities & ESMTP_8BITMIME) == 0 &&
+		!ascii_clean) {
+	      convertmode = _CONVERT_UNKNOWN;
+	      downgrade_headers(startrp, convertmode, verboselog);
+	    }
+	    break;
+	  case 2:		/* MIME, but no C-T-E: -> defaults to 7BIT */
+	  case 1:		/* C-T-E: BASE64  ?? */
+	  case 7:		/* C-T-E: 7BIT */
+	    convertmode = _CONVERT_NONE;
+	    break;
+	  case 8:		/* C-T-E: 8BIT */
+	    if ((ehlo_capabilities & ESMTP_8BITMIME) == 0 &&
+		!ascii_clean) {
+	      convertmode = _CONVERT_QP;
+	      downgrade_headers(startrp, convertmode, verboselog);
+	    }
+	    break;
+	  case 9:		/* C-T-E: Quoted-Printable */
+	    if (force_8bit || (ehlo_capabilities & ESMTP_8BITMIME)) {
+	      /* Force(d) to decode Q-P while transfer.. */
+	      convertmode = _CONVERT_8BIT;
+	      /*  UPGRADE TO 8BIT !  */
+	      qp_to_8bit(startrp);
+	      content_kind = 10;
+	      ascii_clean = 0;
+	    }
+	    break;
+	  default:
+	    /* ???? This should NOT happen! */
+	    break;
+	  } /* switch().. */
+	}
+
+	sprintf(SMTPbuf, "MAIL From:<%s>", startrp->addr->link->user);
 	if (ehlo_capabilities & ESMTP_8BITMIME)
 	  strcat(SMTPbuf, " BODY=8BITMIME");
 	s = SMTPbuf + strlen(SMTPbuf);
-	/* Size estimate is calculated in the  ctlopen()  by
 
+	/* Size estimate is calculated in the  ctlopen()  by
 	   adding msg-body size to the largest known header size,
 	   though excluding possible header and body rewrites.. */
 	if (ehlo_capabilities & ESMTP_SIZEOPT) {
 	  sprintf(s," SIZE=%d",startrp->desc->msgsizeestimate);
 	  s += strlen(s);
 	}
+	/* DSN parameters ... */
 	if ((ehlo_capabilities & ESMTP_DSN) &&
-	    rp->desc->envid != NULL) {
-	  sprintf(s," ENVID=%s",rp->desc->envid);
+	    startrp->desc->envid != NULL) {
+	  sprintf(s," ENVID=%s",startrp->desc->envid);
 	  s += strlen(s);
 	}
 
@@ -553,33 +670,20 @@ deliver(smtpfp, verboselog, startrp, endrp, messagefd, msgoffset)
 	/* Headers are 7-bit stuff -- says MIME specs */
 
 
-	/* Content-Transfer-Encoding: 8BIT ? */
-	content_kind = cte_check(startrp);
-	flag_downgrade = ((content_kind == 8) &&
-			  ((ehlo_capabilities & ESMTP_8BITMIME) == 0));
-	/* If it does not contain MIME headers, and it contains 8-bit
-	   chars, it is in trouble!  Mark the checkup with  -1       */
-	if (content_kind == 0 && (ehlo_capabilities & ESMTP_8BITMIME) == 0)
-	  flag_downgrade = -1;
-	if (flag_downgrade) {
-	  check_7bit_downgrade(messagefd, &flag_downgrade);
-	  if (flag_downgrade & 1) /* Downgrade 'em only if really need it! */
-	    downgrade_headers(startrp, flag_downgrade, verboselog);
-	}
-
 	if (verboselog) {
 	  char **hdrs = *(startrp->newmsgheader);
-	  if (*(startrp->newmsgheadercvt) != NULL && (flag_downgrade & 1))
+	  if (*(startrp->newmsgheadercvt) != NULL &&
+	      convertmode != _CONVERT_NONE)
 	    hdrs = *(startrp->newmsgheadercvt);
 	  fprintf(verboselog,
-		  "Downgraded headers:  ContentKind=%d, flag=%d\n------\n",
-		  content_kind,flag_downgrade);
+		  "Processed headers:  ContentKind=%d, CvtMode=%d\n------\n",
+		  content_kind,convertmode);
 	  while (hdrs && *hdrs)
 	    fprintf(verboselog,"%s\n",*hdrs++);
 	}
 
 	if ((hsize = writeheaders(startrp, smtpfp[1], "\r\n",
-				  flag_downgrade & 1, 0)) < 0) {
+				  convertmode, 0)) < 0) {
 	  for (rp = startrp; rp != endrp; rp = rp->next)
 	    if (rp->status == EX_OK) {
 	      notaryreport(rp->addr->user,notaryacct(r,"*NOTOK*"),
@@ -592,13 +696,17 @@ deliver(smtpfp, verboselog, startrp, endrp, messagefd, msgoffset)
 	  return EX_TEMPFAIL;
 	}
 	
-	/* append message body itself */
-	if (fstat(messagefd, &stbuf) >= 0)
-	  size = hsize + stbuf.st_size - msgoffset;
+	/* Append the message body itself */
+	if (fstat(dp->msgfd, &stbuf) >= 0)
+	  size = hsize + stbuf.st_size - dp->msgbodyoffset;
 	else
 	  size = -hsize;
-	if ((r = appendlet(smtpfp[1], verboselog, messagefd,
-			   hsize, size, flag_downgrade)) != EX_OK) {
+
+	if (lseek(dp->msgfd, dp->msgbodyoffset, L_SET) < 0L)
+	  warning("Cannot seek to message body! (%m)", (char *)NULL);
+
+	if ((r = appendlet(dp, smtpfp[1], verboselog,
+			   hsize, size, convertmode)) != EX_OK) {
 	  for (rp = startrp; rp != endrp; rp = rp->next)
 	    if (rp->status == EX_OK) {
 	      notaryreport(rp->addr->user, notaryacct(r,"*NOTOK*"),
@@ -608,7 +716,7 @@ deliver(smtpfp, verboselog, startrp, endrp, messagefd, msgoffset)
 	  /*
 	   * The failure was due to an I/O error either in the local
 	   * message file, because the socket got torn down under our
-	   * feet, or because the other end is out of sync.  In case 1
+	   * feet, or because the other end is out of sync.  In cases 1
 	   * and 3 we want to try to convince the remote to abort the
 	   * transaction.  Issuing a RSET is empirically a Good Thing
 	   * in this situation, as suggested by Matti Aarnio.  If we
@@ -657,164 +765,210 @@ deliver(smtpfp, verboselog, startrp, endrp, messagefd, msgoffset)
  * appendlet - append letter to file pointed at by fd
  */
 
+#ifndef	USE_MMAP
 static char let_buffer[BUFSIZ*8];
-
+#endif
 
 int
-appendlet(fp, verboselog, mfd, hsize, msize, downgrade)
+appendlet(dp, fp, verboselog, hsize, msize, convertmode)
+	struct ctldesc *dp;
 	FILE *fp, *verboselog;
-	int mfd, hsize, msize, downgrade;
+	int hsize, msize, convertmode;
 {
-	/* downgrade&1  is non-zero  when MIME-tag Content-Transfer-Encoding:
-	   is altered from  8BIT to  QUOTED-PRINTABLE.   When   downgrade&2 is
-	   non-zero, alteration is from 8BIT to 7BIT.
+	/* `convertmode' controls the behaviour of the message conversion:
+	     _CONVERT_NONE (0): send as is
+	     _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
+	     _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
+	     _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
 	 */
 
 	register int i;
 	register int bufferfull;
+	int position = dp->msgbodyoffset;
 	int lastwasnl;
 
-	writebuf(fp, (char *)NULL, 0, 0);	/* magic init. */
+	writebuf(fp, (char *)NULL, 0);	/* magic init. */
 
+#ifdef TIMEOUT
+	alarm(1*60*60); /* Timeout in 1 hour per let_buffer full..
+			   BUFSIZ/h is damn slow.. It might not be that
+			   slow for some monster-files, but.. */
+	gotalarm = 0;
+#endif
+#ifndef	USE_MMAP
 	/* can we use cache of message body data */
-	if (readalready != 0) {
-	  if (writebuf(fp, let_buffer, readalready, downgrade) != readalready)
+	if (convertmode == _CONVERT_NONE &&
+	    readalready != 0) {
+	  if (writebuf(fp, let_buffer, readalready) != readalready)
+#ifdef TIMEOUT
+	    if (gotalarm) {
+	      sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
+		      hsize, msize, (hsize*100+msize/2)/msize);
+	      return EOF;
+	    }
+#endif
 	    return EX_IOERR;
 	  if (let_buffer[readalready-1] != '\n')
-	    writebuf(fp, "\n", 1, 0);
+	    writebuf(fp, "\n", 1);
 	  if (statusreport)
 	    report("DATA %d (100%%)", readalready+hsize);
 	  return EX_OK;
 	}
-
+#endif
 	/* we are assumed to be positioned properly at start of message body */
 	bufferfull = 0;
 	lastwasnl = 1;	/* we are guaranteed to have a \n after the header */
-#ifdef TIMEOUT
-	alarm(1*60*60); /* Timeout in 1 hour per let_buffer full... BUFSIZ/h is
-			   damn slow.. */
-	gotalarm = 0;
-#endif
-	while ((i = read(mfd, let_buffer, sizeof let_buffer)) != 0) {
-	  if (i < 0) {
-	    notaryreport(NULL,NULL,"500 (Read error from message file!?)");
-	    strcpy(remotemsg, "read error from message file");
-	    if (verboselog)
-	      fprintf(verboselog,
-		      "source read error! (%s)  DATA %d/%d (%d%%)",
-		      strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
-	    return EX_IOERR;
-	  }
-	  lastwasnl = (let_buffer[i-1] == '\n');
-	  if (writebuf(fp, let_buffer, i, downgrade) != i) {
-#ifdef TIMEOUT
-	    if (gotalarm) {
-	      sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
-		      hsize, msize, (hsize*100+msize/2)/msize);
-	      notaryreport(NULL,NULL,remotemsg);
-	      if (verboselog)
-		fprintf(verboselog,"%s\n",remotemsg);
-	      if (bufferfull > 1) readalready = 0;
-	      return EX_TEMPFAIL;
+	if (convertmode == _CONVERT_NONE) {
+#ifdef	USE_MMAP
+	  char *let_buffer = dp->let_buffer + dp->msgbodyoffset;
+	  i = dp->let_end - dp->let_buffer - dp->msgbodyoffset;
+#else /* !USE_MMAP */
+	  while (1) {
+	    if ((i = read(dp->msgfd, let_buffer, sizeof(let_buffer))) == 0)
+	      break;
+	    if (i < 0) {
+	      strcpy(remotemsg, "500 (Read error from message file!?)");
+	      return EOF;
 	    }
+#endif /* !USE_MMAP */
+	    lastwasnl = (let_buffer[i-1] == '\n');
+	    if (writebuf(fp, let_buffer, i) != i) {
+#ifdef TIMEOUT
+	      if (gotalarm) {
+		sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
+			hsize, msize, (hsize*100+msize/2)/msize);
+		return EOF;
+	      }
 #endif
-	    sprintf(remotemsg,
-		    "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
-		    strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
-	    notaryreport(NULL,NULL,remotemsg);
-	    if (verboselog)
-	      fprintf(verboselog,"%s\n",remotemsg);
+	      sprintf(remotemsg,
+		      "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
+		      strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
+	      return EOF;
+	    }
+	    if (statusreport) {
+	      hsize += i;
+	      report("DATA %d/%d (%d%%)",
+		     hsize, msize, (hsize*100+msize/2)/msize);
+	    }
+#ifndef USE_MMAP
+	    bufferfull++;
+	    readalready = i;
+	  }
+#endif
+	} else {
+	  /* Various esoteric conversion modes..
+	     We are better to feed writemimeline() with LINES
+	     instead of blocks of data.. */
+#ifndef	USE_MMAP /* Classical way to read in things */
+	  char iobuf[BUFSIZ];
+	  FILE *mfp = fdopen(dp->msgfd,"r");
+	  setvbuf(mfp, iobuf, _IOFBF, sizeof(iobuf));
+#define MFPCLOSE i = dup(dp->msgfd); fclose(mfp); dup2(i,dp->msgfd); close(i);
+	  bufferfull = 0;
+#else /* USE_MMAP */
+	  char *s2, *s = dp->let_buffer + dp->msgbodyoffset;
 
-	    if (bufferfull > 1) readalready = 0;
-	    return EX_IOERR;
+#define MFPCLOSE
+#endif
+	  writemimeline(fp, (char *)NULL, 0, 0);
+
+	  /* we are assuming to be positioned properly
+	     at the start of the message body */
+#ifndef	USE_MMAP
+	  readalready = 0;
+#endif
+	  lastwasnl = 0;
+	  for (;;) {
+#ifndef USE_MMAP
+	    if ((i = cfgets(let_buffer, sizeof(let_buffer), mfp)) == EOF)
+	      break;
+	    /* It MAY be malformed -- if it has a BUFSIZ*8 length
+	       line in it, IT CAN'T BE MIME  :-/		*/
+	    lastwasnl = (let_buffer[i-1] == '\n');
+#else /* USE_MMAP */
+	    char *let_buffer = s, *s2 = s;
+	    if (s >= dp->let_end) break; /* EOF */
+	    i = 0;
+	    while (s2 < dp->let_end && *s2 != '\n')
+	      ++s2, ++i;
+	    if ((lastwasnl = (*s2 == '\n')))
+	      ++s2, ++i;
+	    s = s2;
+#endif
+	    /* XX: Detect multiparts !! */
+
+	    /* Ok, write the line */
+	    if (writemimeline(fp, let_buffer, i, convertmode) != i) {
+	      sprintf(remotemsg,
+		      "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
+		      strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
+	      MFPCLOSE
+	      return EOF;
+	    }
+#ifdef USE_MMAP
+	    s = s2; /* Advance one linefull.. */
+#endif
 	  }
-	  if (statusreport) {
-	    hsize += i;
-	    report("DATA %d/%d (%d%%)",
-		   hsize, msize, (hsize*100+msize/2)/msize);
+#ifndef	USE_MMAP
+	  if (i == EOF && !feof(mfp)) {
+	    strcpy(remotemsg, "500 (Read error from message file!?)");
+	    MFPCLOSE
+	    return EOF;
 	  }
-	  readalready = i;
-	  bufferfull++;
-	}
+	  MFPCLOSE
+#endif
+	} /* ... end of conversion modes */
+
 	/* we must make sure the last thing we transmit is a CRLF sequence */
 	if (!lastwasnl)
-	  if (writebuf(fp, "\n", 1, downgrade) != 1) {
+	  if (writebuf(fp, "\n", 1) != 1) {
 #ifdef TIMEOUT
 	    if (gotalarm) {
 	      sprintf(remotemsg,"500 (let_buffer write timeout!  DATA %d/%d [%d%%])",
 		      hsize, msize, (hsize*100+msize/2)/msize);
-	      notaryreport(NULL,NULL,remotemsg);
-	      if (verboselog)
-		fprintf(verboselog,"%s\n",remotemsg);
-	      if (bufferfull > 1) readalready = 0;
-	      return EX_TEMPFAIL;
+	      return EOF;
 	    }
 #endif
 	    sprintf(remotemsg,
 		    "500 (let_buffer write IO-error! [%s] DATA %d/%d [%d%%])",
 		    strerror(errno), hsize, msize, (hsize*100+msize/2)/msize);
-	    notaryreport(NULL,NULL,remotemsg);
-	    if (verboselog)
-	      fprintf(verboselog,"%s\n",remotemsg);
-
+#ifndef	USE_MMAP
 	    if (bufferfull > 1) readalready = 0;
-	    return EX_IOERR;
+#endif
+	    return EOF;
 	  }
 
+#ifndef USE_MMAP
 	if (bufferfull > 1)	/* not all in memory, need to reread */
 	  readalready = 0;
-	return EX_OK;
+#endif
+	return 0;
 }
 
 /*
  * Writebuf() is like write(), except all '\n' are converted to "\r\n"
  * (CRLF), and the sequence "\n." is converted to "\r\n..".
+ * writebuf() has a cousin: writemimeline(), which does some more esoteric
+ * conversions on flight..
  */
-
 int
-writebuf(fp, buf, len, downgrade)
+writebuf(fp, buf, len)
 	FILE *fp;
 	char *buf;
 	int len;
-	int downgrade;
 {
 	register char *cp;
 	register int n;
-	int tlen;
-	register int state, column;
-	char *i2h = "0123456789ABCDEF";
+	register int state;
 	static int save = 0;
-	static int col  = 0;
-	int qp_conv = (downgrade & 1) /* && !(downgrade & 2) */ ;
 
 	if (buf == NULL) {		/* magic initialization */
 	  save = 1;
-	  col  = 0;
 	  return 0;
 	}
 	state  = save;
-	column = col;
-	for (cp = buf, n = len, tlen = 0; n > 0; --n, ++cp) {
+	for (cp = buf, n = len; n > 0; --n, ++cp) {
 	  int c = *(unsigned char *)cp;
-	  if (qp_conv && column > 70 && c != '\n') {
-	    putc('=',fp);
-	    putc('\r',fp);
-	    putc('\n',fp);
-	    /* if (verboselog) fprintf(verboselog,"=\n"); */
-	    column = 0;
-	    state = 1;
-	  }
-	  if (qp_conv &&
-	      (c == '='  ||  c > 126 ||  (c != '\n' && c < 32))) {
-
-	    putc('=',fp);
-	    putc(i2h[(c >> 4) & 15],fp);
-	    putc(i2h[(c)      & 15],fp);
-	    column += 3;
-	    /* if (verboselog) fprintf(verboselog,"=%02X",c); */
-	    ++tlen;
-	    continue;
-	  }
 
 	  if (state && c != '\n') {
 	    state = 0;
@@ -824,7 +978,6 @@ writebuf(fp, buf, len, downgrade)
 		strcpy(remotemsg, "write error 1");
 		return EOF;
 	      }
-	      ++column;
 	      /* if (verboselog) fprintf(verboselog,".."); */
 	    } else {
 	      if (putc(c, fp) == EOF) {
@@ -834,8 +987,6 @@ writebuf(fp, buf, len, downgrade)
 	      }
 	      /* if (verboselog) putc(c, verboselog); */
 	    }
-	    ++column;
-	    ++tlen;
 	  } else if (c == '\n') {
 	    if (putc('\r', fp) == EOF || putc(c, fp) == EOF) {
 	      notaryreport(NULL,NULL,"500 (body write error, 3)");
@@ -843,8 +994,6 @@ writebuf(fp, buf, len, downgrade)
 	      return EOF;
 	    }
 	    /* if (verboselog) putc(c, verboselog); */
-	    ++tlen;
-	    column = 0;
 	    state = 1;
 	  } else {
 	    if (putc(c, fp) == EOF) {
@@ -853,21 +1002,137 @@ writebuf(fp, buf, len, downgrade)
 	      return EOF;
 	    }
 	    /* if (verboselog) putc(c, verboselog); */
-	    ++column;
-	    ++tlen;
 	  }
 	}
-	col  = column;
 	save = state;
-	return tlen;
+	return len;
 }
 
+
+int
+writemimeline(fp, buf, len, convertmode)
+	FILE *fp;
+	char *buf;
+	int len;
+	int convertmode;
+{
+	register char *cp;
+	register int n;
+	static int column;
+	char *i2h = "0123456789ABCDEF";
+	int qp_chrs = 0;
+	int qp_val = 0;
+	int qp_conv;
+
+	/* `convertmode' controls the behaviour of the message conversion:
+	     _CONVERT_NONE (0): send as is
+	     _CONVERT_QP   (1): Convert 8-bit chars to QUOTED-PRINTABLE
+	     _CONVERT_8BIT (2): Convert QP-encoded chars to 8-bit
+	     _CONVERT_UNKNOWN (3): Turn message to charset=UNKNOWN-8BIT, Q-P..
+	 */
+
+	if (buf == NULL) {
+	  column = -1;
+	  return 0;
+	}
+
+	qp_conv = (convertmode == _CONVERT_QP ||
+		   convertmode == _CONVERT_UNKNOWN);
+
+	if (buf == NULL) {		/* magic initialization */
+	  /* No magics here.. we are linemode.. */
+	  return 0;
+	}
+	for (cp = buf, n = len; n > 0; --n, ++cp) {
+	  int c = *(unsigned char *)cp;
+	  ++column;
+	  if (convertmode == _CONVERT_8BIT) {
+	    if (qp_chrs == 0 && c == '=') {
+	      qp_val = 0;
+	      qp_chrs = 2;
+	      continue;
+	    }
+	    if (qp_chrs != 0) {
+	      if (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
+		n = 0;	/* We have the line-end wrapper mode */
+		continue; /* It should NEVER be present except at the end
+			     of the line, thus we are safe to do this ? */
+	      }
+	      --column;
+	      if ((c >= '0' && c <= '9') ||
+		  (c >= 'a' && c <= 'f') ||
+		  (c >= 'A' && c <= 'F')) {
+		/* The first char was HEX digit, assume the second one
+		   to be also, and convert all three (=XX) to be a char
+		   of given value.. */
+		if (c >= 'a') c -= ('a' - 'A');
+		if (c > '9') c -= ('A' - '9' - 1);
+		qp_val <<= 4;
+		qp_val |= (c & 0x0F);
+	      }
+	      --qp_chrs;
+	      if (qp_chrs == 0)
+		c = qp_val;
+	      else
+		continue;
+	    }
+	  } else {
+	    if (qp_conv && column > 70 && c != '\n') {
+	      putc('=',fp);
+	      putc('\r',fp);
+	      putc('\n',fp);
+	      /* if (verboselog) fprintf(verboselog,"=\n"); */
+	      column = -1;
+	    }
+	    if (qp_conv &&
+		(c == '='  ||  c > 126 ||  (c != '\n' && c < 32))) {
+
+	      putc('=',fp);
+	      putc(i2h[(c >> 4) & 15],fp);
+	      putc(i2h[(c)      & 15],fp);
+	      column += 3;
+	      /* if (verboselog) fprintf(verboselog,"=%02X",c); */
+	      continue;
+	    }
+	  }
+
+	  if (column == 0 && c == '.') {
+	    if (putc(c, fp) == EOF) {
+	      notaryreport(NULL,NULL,"500 (body write error, 1)");
+	      strcpy(remotemsg, "write error 1");
+	      return EOF;
+	    }
+	    /* if (verboselog) fprintf(verboselog,".."); */
+	  }
+
+	  if (c == '\n') {
+	    if (putc('\r', fp) == EOF) {
+	      notaryreport(NULL,NULL,"500 (body write error, 3)");
+	      strcpy(remotemsg, "write error 3");
+	      return EOF;
+	    }
+	    /* if (verboselog) putc(c, verboselog); */
+	    column = -1;
+	  }
+	  if (putc(c, fp) == EOF) {
+	    notaryreport(NULL,NULL,"500 (body write error, 4)");
+	    strcpy(remotemsg, "write error 4");
+	    return EOF;
+	  }
+	  /* if (verboselog) putc(c, verboselog); */
+
+	}
+	return len;
+}
+
+
 int
 ehlo_check(buf)
 char *buf;
 {
-
-	if (strncmp(buf,"8BITMIME",8)==0) {
+	char *r = strchr(buf,'\r');
+	if (r != NULL) *r = 0;
+	if (strcmp(buf,"8BITMIME")==0) {
 	  ehlo_capabilities |= ESMTP_8BITMIME;
 	} else if (strcmp(buf,"XDSN")==0 ||
 		   strcmp(buf,"DSN")==0) {
@@ -964,6 +1229,7 @@ smtpconn(host, fparr, verboselog)
 	struct mxdata mxh[MAXFORWARDERS];
 	int	mxcnt = 0;
 
+	hp = NULL;
 #ifdef	BIND
 	h_errno = 0;
 #endif	/* BIND */
@@ -1260,10 +1526,13 @@ vcsetup(ipaddrcp, ipaddrlen, fdp, hostname, verboselog)
 	    return EX_TEMPFAIL;
 	  }
 	}
-	bcopy(ipaddrcp, (char *)&sad.sin_addr, ipaddrlen);
+	memcpy((char *)&sad.sin_addr, ipaddrcp, ipaddrlen);
 	sad.sin_port = htons((u_short)IPPORT_SMTP);
+/* setreuid(0,first_uid);
+   if(verboselog) fprintf(verboselog,"setreuid: first_uid=%d, ruid=%d, euid=%d\n",first_uid,getuid(),geteuid()); */
 	if (connect(s, (struct sockaddr *)&sad, sizeof sad) >= 0) {
 	  int on = 1;
+/* setreuid(0,0); */
 	  *fdp = s;
 #if	BSD >= 43
 	  setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, &on, sizeof on);
@@ -1274,6 +1543,7 @@ vcsetup(ipaddrcp, ipaddrlen, fdp, hostname, verboselog)
 	    fprintf(stderr, "connected!\n");
 	  return EX_OK;
 	}
+/* setreuid(0,0); */
 
 	errnosave = errno;
 	se = strerror(errno);
@@ -1462,7 +1732,7 @@ smtpwrite(smtpfp, verboselog, saverpt, strbuf)
 		}
 
 		if (s + 1 < cp)
-		  bcopy(s+1, buf, cp-s-1);
+		  memcpy(buf, s+1, cp-s-1);
 		cp = buf + (cp-s-1);
 		s = buf;
 		--s;		/* incremented in for() stmt */
@@ -1510,7 +1780,8 @@ smtpwrite(smtpfp, verboselog, saverpt, strbuf)
 	      {
 		se = strerror(errno);
 		sprintf(remotemsg, "500 (Error on %sSMTP responce read: %s, Cmd: %s)",
-			strbuf == NULL ? "initial ":"", se, strbuf);
+			strbuf == NULL ? "initial ":"", se,
+			strbuf == NULL ? "(null)" : strbuf);
 		notaryreport(NULL,NULL,remotemsg);
 	      }
 	    dflag = 0;
@@ -1688,7 +1959,7 @@ stashmyaddresses(host)
 	if ((myaddrs = (char *)malloc(nmyaddrs * hp->h_length)) == NULL)
 	  return;
 	for (hp_init(hp); *hp_getaddr() != NULL; hp_nextaddr())
-	  bcopy(*hp_getaddr(), &myaddrs[(--naddrs)*hp->h_length], hp->h_length);
+	  memcpy(&myaddrs[(--naddrs)*hp->h_length], *hp_getaddr(), hp->h_length);
 }
 
 int
@@ -2174,11 +2445,19 @@ matchroutermxes(spec_host, ap)
 
 
 /* When data is clean 7-BIT, do:  *flag_ptr = (*flag_ptr) << 1  */
-void
-check_7bit_downgrade(mfd, flag_ptr)
-int mfd;
-int *flag_ptr;
+int 
+check_7bit_cleanness(dp)
+struct ctldesc *dp;
 {
+#ifdef	USE_MMAP
+	/* With MMAP()ed spool file it is sweet and simple.. */
+	register char *s = dp->let_buffer + dp->msgbodyoffset;
+	while (s < dp->let_end)
+	  if (128 & *s) return 0;
+	  else ++s;
+	return 1;
+#else /* !USE_MMAP */
+
 	register int i;
 	register int bufferfull;
 	int lastwasnl;
@@ -2188,7 +2467,7 @@ int *flag_ptr;
 	if (readalready != 0) {
 	  for (i=0; i<readalready; ++i)
 	    if (128 & (let_buffer[i]))
-	      return;		/* Not clean ! */
+	      return 0;		/* Not clean ! */
 	}
 
 	/* we are assumed to be positioned properly at start of message body */
@@ -2200,7 +2479,7 @@ int *flag_ptr;
 	  if (i < 0) {
 	    /* ERROR ?!?!? */
 	    if (bufferfull > 1) readalready = 0;
-	    return;
+	    return 0;
 	  }
 	  lastwasnl = (let_buffer[i-1] == '\n');
 	  readalready = i;
@@ -2209,7 +2488,7 @@ int *flag_ptr;
 	    if (128 & (let_buffer[i])) {
 	      lseek(mfd,mfd_pos,0);
 	      if (bufferfull > 1) readalready = 0;
-	      return;		/* Not clean ! */
+	      return 0;		/* Not clean ! */
 	    }
 	}
 	/* Got to EOF, and still it is clean 7-BIT! */
@@ -2218,7 +2497,8 @@ int *flag_ptr;
 	  readalready = 0;
 
 	/* Set indication! */
-	*flag_ptr = (*flag_ptr) << 1;
+	return 1;
+#endif /* !USE_MMAP */
 }
 
 void
@@ -2255,3 +2535,13 @@ char *smtpline;
 	*s = 0;
 	notaryreport(NULL,NULL,str);
 }
+
+
+void getdaemon()
+{
+	struct passwd *pw = getpwnam("daemon");
+	if (!pw) pw = getpwnam("uucp");
+
+	if (!pw) daemon_uid = 0; /* Let it be root, if nothing else */
+	else     daemon_uid = pw->pw_uid;
+}
